from abc import ABC, abstractmethod
from typing import Dict, Optional, Tuple

import numpy as np

from ttfemesh.basis import BilinearBasis, TensorProductBasis
from ttfemesh.domain import Domain, Quad
from ttfemesh.domain.subdomain_connection import CurveConnection2D, VertexConnection2D
from ttfemesh.mesh.mesh_utils import side_concatenation_tt, vertex_concatenation_tt
from ttfemesh.mesh.subdomain_mesh import QuadMesh, SubdomainMesh, SubdomainMesh2D
from ttfemesh.quadrature.quadrature import QuadratureRule
from ttfemesh.tt_tools.tensor_cross import TTCrossConfig
from ttfemesh.types import BoundarySide2D, BoundaryVertex2D, TensorTrain


class DomainMesh(ABC):
    """
    DomainMesh base class that ties together the domain, subdomain meshes, and basis functions.
    It provides an interface for the element jacobians of the subdomains,
    the element to global index maps, the boundary masks and the concatenation maps.
    """

    def __init__(
        self,
        domain: Domain,
        quadrature_rule: QuadratureRule,
        mesh_size_exponent: int,
        basis: TensorProductBasis,
        tt_cross_config: Optional[TTCrossConfig] = None,
    ):
        """
        Initialize a DomainMesh.

        Args:
            domain (Domain): The domain containing subdomains and their connections.
            quadrature_rule (QuadratureRule): Quadrature rule for integration.
            mesh_size_exponent (int): Discretization size exponent.
            basis (TensorProductBasis): The basis functions for the domain.
            cross_config (Optional[TTCrossConfig]):
                Optional configuration for tensor cross approximation.
                If None, the default configuration is used.
        """

        self.quadrature_rule = quadrature_rule
        self.mesh_size_exponent = mesh_size_exponent
        self.domain = domain
        self.basis = basis
        self._tt_cross_config = tt_cross_config
        self.subdomain_meshes = self._create_subdomain_meshes()

    @abstractmethod
    def _create_subdomain_meshes(self):  # pragma: no cover
        num_subdomains = self.domain.num_subdomains
        subdomain_meshes = []
        for i in range(num_subdomains):
            subdomain = self.domain.get_subdomain(i)
            mesh_size_exponent = self.mesh_size_exponent
            quadrature_rule = self.quadrature_rule
            tt_cross_config = self._tt_cross_config
            subdomain_mesh = SubdomainMesh(
                subdomain, quadrature_rule, mesh_size_exponent, tt_cross_config
            )
            subdomain_meshes.append(subdomain_mesh)

        return subdomain_meshes

    def get_subdomain_mesh(self, subdomain_index: int) -> SubdomainMesh:  # noqa
        """
        Get the SubdomainMesh for a subdomain.

        Args:
            subdomain_index (int): The index of the subdomain.

        Returns:
            SubdomainMesh: The SubdomainMesh for the subdomain.

        Raises:
            ValueError: If the subdomain index is invalid.
        """
        self._validate_subdomain_index(subdomain_index)
        return self.subdomain_meshes[subdomain_index]

    def get_element2global_index_map(self) -> np.ndarray:  # noqa
        """
        Get the TT-representation of transformations mapping from element index to global basis
        function index for all reference basis functions on a reference element in a subdomain.
        This map depends on the type of basis functions used and
        the discretization size of the subdomain.

        Args:
            subdomain_index (int): The index of the subdomain.

        Returns:
            np.ndarray: A matrix of TT-representations, i.e.,
                each element of the matrix is a TT-vector.
                Indexing of the matrix corresponds to the
                reference basis function indexing.
                For example, for a bilinear basis in 2D, the indexing is:
                (i, j) where i and j are the indices of the basis functions in x and y direction
                Specifically, (0, 0), (1, 0), (0, 1) and (1, 1),
                representing the four basis functions
                corresponding to the four vertices of the reference element:
                lower left, lower right, upper right and upper left, respectively.
                See also the documentation for the chosen Basis class.
        """
        mesh_size_exponent = self.mesh_size_exponent
        ttmaps = self.basis.get_all_element2global_ttmaps(mesh_size_exponent)

        return ttmaps

    def get_dirichlet_masks(self) -> Dict[int, TensorTrain]:  # noqa
        """
        Get the dirichlet boundary masks.

        Returns:
            Dict[TensorTrain]: A dictionary where the keys are subdomain indices,
                and the values are TT-representations of the boundary masks.
        """

        boundary_condition = self.domain.boundary_condition
        if boundary_condition is None:
            print("No boundary condition specified.")
            return None

        grouped = boundary_condition.group_by_subdomain()
        boundary_masks = {}
        for subdomain_index, curve_indices in grouped.items():
            sides = [BoundarySide2D(i) for i in curve_indices]
            boundary_mask = self.basis.get_dirichlet_mask(self.mesh_size_exponent, *sides)
            boundary_masks[subdomain_index] = boundary_mask

        return boundary_masks

    @abstractmethod  # noqa # pragma: no cover
    def get_concatenation_maps(self) -> Dict[Tuple[int, int], TensorTrain]:
        """
        Get the TT-representations of the concatenation maps for all pairs of connected subdomains.

        Returns:
            Dict[Tuple[int, int], TensorTrain]:
                A dictionary where the keys are pairs of subdomain indices,
                and the values are TT-representations of the concatenation maps.
        """
        pass

    def _validate_subdomain_index(self, subdomain_index: int):
        if subdomain_index < 0 or subdomain_index >= self.domain.num_subdomains:
            raise ValueError(
                f"Invalid subdomain index: {subdomain_index}. "
                f"Valid indices are in the range [0, {self.domain.num_subdomains})."
            )

    def __repr__(self) -> str:
        return (
            f"DomainMesh(domain={self.domain}, "
            f"mesh_size_exponent={self.mesh_size_exponent}, "
            f"quadrature_rule={self.quadrature_rule}, "
            f"basis={self.basis})"
        )


class DomainMesh2D(DomainMesh):
    """Mesh for 2D domains. This is essentially a factory for SubdomainMesh2D objects."""

    def _create_subdomain_meshes(self):
        num_subdomains = self.domain.num_subdomains
        subdomain_meshes = []
        for i in range(num_subdomains):
            subdomain = self.domain.get_subdomain(i)
            mesh_size_exponent = self.mesh_size_exponent
            quadrature_rule = self.quadrature_rule
            tt_cross_config = self._tt_cross_config
            if isinstance(subdomain, Quad):
                subdomain_mesh = QuadMesh(
                    quad=subdomain,
                    quadrature_rule=quadrature_rule,
                    mesh_size_exponent=mesh_size_exponent,
                    tt_cross_config=tt_cross_config,
                )
            else:
                subdomain_mesh = SubdomainMesh2D(
                    subdomain=subdomain,
                    quadrature_rule=quadrature_rule,
                    mesh_size_exponent=mesh_size_exponent,
                    tt_cross_config=tt_cross_config,
                )
            subdomain_meshes.append(subdomain_mesh)

        return subdomain_meshes

    def __repr__(self) -> str:
        return (
            f"DomainMesh2D(domain={self.domain}, "
            f"mesh_size_exponent={self.mesh_size_exponent}, "
            f"quadrature_rule={self.quadrature_rule}, "
            f"basis={self.basis})"
        )


class DomainBilinearMesh2D(DomainMesh2D):
    """
    Mesh for 2D domains with bilinear basis functions.
    This implementation of the concatenation maps works only for bilinear basis functions.

    Example:
    >>> from ttfemesh.domain import RectangleFactory, CurveConnection2D, VertexConnection2D
    >>> from ttfemesh.domain import DirichletBoundary2D, Domain2D
    >>> from ttfemesh.quadrature import GaussLegendre2D
    >>> from ttfemesh.mesh import DomainBilinearMesh2D

    >>> lower_left = (0, 0)
    >>> upper_right = (2, 1)
    >>> rectangle1 = RectangleFactory.create(lower_left, upper_right)

    >>> lower_left = (2, 0)
    >>> upper_right = (3, 1)
    >>> rectangle2 = RectangleFactory.create(lower_left, upper_right)

    >>> lower_left = (-2, 1)
    >>> upper_right = (0, 2)
    >>> rectangle3 = RectangleFactory.create(lower_left, upper_right)

    >>> domain_idxs = [0, 1]
    >>> curve_idxs = [1, 3]
    >>> edge = CurveConnection2D(domain_idxs, curve_idxs)

    >>> vertex_idxs = [(0, 3, "start"), (2, 0, "end")]
    >>> vertex = VertexConnection2D(vertex_idxs)

    >>> bc = DirichletBoundary2D([(1, 1), (2, 3)])

    >>> domain = Domain2D([rectangle1, rectangle2, rectangle3], [edge, vertex], bc)

    >>> quadrature_rule = GaussLegendre2D()
    >>> mesh_size_exponent = 3
    >>> domain_mesh = DomainBilinearMesh2D(domain, quadrature_rule, mesh_size_exponent)
    >>> print(domain_mesh)
    """

    def __init__(
        self,
        domain: Domain,
        quadrature_rule: QuadratureRule,
        mesh_size_exponent: int,
        tt_cross_config: Optional[TTCrossConfig] = None,
    ):
        basis = BilinearBasis()
        super().__init__(domain, quadrature_rule, mesh_size_exponent, basis, tt_cross_config)

    def get_concatenation_maps(  # noqa
        self,
    ) -> Dict[Tuple[int, int], Tuple[TensorTrain, TensorTrain, TensorTrain]]:
        """
        Get the TT-representations of the concatenation maps for all pairs of connected subdomains.
        See Section 5 of arXiv:1802.02839 for details.
        Pmp describes which nodes in the left domain are connected
        to which nodes in the right domain,
        Pmm describes which nodes in the left domain are to be connected,
        Ppp describes which nodes in the right domain are to be connected.

        Returns:
            Dict[Tuple[int, int], Tuple[TensorTrain, TensorTrain, TensorTrain]]:
                A dictionary where the keys are pairs of subdomain indices,
                and the values are tuples TT-representations of the connectivity maps.

        Raises:
            ValueError: If the connection type is not supported.
        """

        connections = self.domain.get_connections()
        concatenation_maps = {}
        for connection in connections:
            if isinstance(connection, VertexConnection2D):
                for (
                    (subdidx1, subdidx2),
                    (curveidx1, curveidx2),
                    (pos1, pos2),
                ) in connection.get_connection_pairs():

                    offset1 = 0 if pos1 == "start" else 1
                    vertex_idx1 = (curveidx1 + offset1) % 4

                    offset2 = 0 if pos2 == "start" else 1
                    vertex_idx2 = (curveidx2 + offset2) % 4

                    vertex1 = BoundaryVertex2D(vertex_idx1)
                    vertex2 = BoundaryVertex2D(vertex_idx2)

                    tt_connectivity = vertex_concatenation_tt(
                        vertex1, vertex2, self.mesh_size_exponent
                    )
                    concatenation_maps[(subdidx1, subdidx2)] = tt_connectivity

            elif isinstance(connection, CurveConnection2D):
                subdidx1, subdidx2 = connection.subdomains_indices
                curveidx1, curveidx2 = connection.curve_indices
                side1 = BoundarySide2D(curveidx1)
                side2 = BoundarySide2D(curveidx2)
                tt_connectivity = side_concatenation_tt(side1, side2, self.mesh_size_exponent)
                concatenation_maps[(subdidx1, subdidx2)] = tt_connectivity

            else:
                raise ValueError(f"Unsupported connection type: {type(connection)}.")

        return concatenation_maps

    def __repr__(self):
        return (
            f"DomainBilinearMesh2D(domain={self.domain}, "
            f"mesh_size_exponent={self.mesh_size_exponent}, "
            f"quadrature_rule={self.quadrature_rule}, "
            f"basis={self.basis})"
        )



import copy

import torch

from ttfemesh.types import TensorTrain


def zorder_kron(left: TensorTrain, right: TensorTrain) -> TensorTrain:
    """
    Compute the Kronecker product of two TT-tensors using the Z-ordering
    (a.k.a. *transposed* or *level-wise* ordering).
    The ordering is column-major, i.e., for every index z in the resulting tensor,
    the relationship to index (i, j) in the left and right tensors is: z = i + j * 2.
    This is consistent with the little-endian ordering of the binary index
    commonly used in QTT literature.

    Args:
        left (TensorTrain): Left TT-tensor.
        right (TensorTrain): Right TT-tensor.

    Returns:
        TensorTrain resulting from the Kronecker product of left and right.

    Raises:
        ValueError: If the TT-length of left and right tensors are not equal.
    """

    cores_left = left.cores
    cores_right = right.cores

    if len(cores_left) != len(cores_right):
        raise ValueError(
            f"TT-length of left ({len(cores_left)})"
            f" and right ({len(cores_right)}) tensors must be equal."
        )

    cores = [torch.kron(b, a) for a, b in zip(cores_left, cores_right)]

    return TensorTrain(cores)


def zorder_linfunc2d(c: float, cx: float, X: TensorTrain, cy: float, Y: TensorTrain) -> TensorTrain:
    """
    Compute the linear combination of two TT-tensors using the Z-ordering
    (a.k.a. *transposed* or *level-wise* ordering):
    c + cx * X + cy * Y.
    Note that both X and Y must be of the same TT-length and have at least rank 2.

    Args:
        c (float): Scalar constant.
        cx (float): Scalar factor for X.
        X (TensorTrain): First TT-tensor.
        cy (float): Scalar factor for Y.
        Y (TensorTrain): Second TT-tensor.

    Returns:
        TensorTrain resulting from the linear combination of c, X, and Y.

    Raises:
        ValueError: If X and Y have different TT-lengths or ranks smaller than 2.
    """

    X_cores, Y_cores = X.cores, Y.cores
    if len(X_cores) != len(Y_cores):
        raise ValueError("X and Y must have the same TT-length.")

    ranks_X, ranks_Y = X.R, Y.R
    for i in range(1, len(ranks_X) - 1):
        if ranks_X[i] == 1:
            raise ValueError("X must have at least rank 2.")
        if ranks_Y[i] == 1:
            raise ValueError("Y must have at least rank 2.")

    result = copy.deepcopy(X_cores)
    result[0][:, :, 0] = cx * X_cores[0][:, :, 0] + cy * Y_cores[0][:, :, 0]
    result[-1][1, :, :] = cx * X_cores[-1][1, :, :] + (c + cy * Y_cores[-1][1, :, :])

    d = len(result)
    for k in range(1, d - 1):
        result[k][1, :, 0] = cx * X_cores[k][1, :, 0] + cy * Y_cores[k][1, :, 0]

    return TensorTrain(result)


# Aliases
transpose_kron = zorder_kron
levelwise_kron = zorder_kron



from typing import Union

import numpy as np


def ensure_1d(array_or_scalar: Union[np.ndarray, float]) -> np.ndarray:
    """
    Ensure the input is a 1D array.

    Args:
        array_or_scalar (Union[np.ndarray, float]): Input that can be a 1D array
            of shape (num_pts,) or a scalar.

    Returns:
        np.ndarray: A 1D array. If input is a scalar, it is converted to a 1D array
        with a single element.

    Raises:
        ValueError: If input is not a scalar or a 1D array.
    """

    if np.isscalar(array_or_scalar):
        return np.array([array_or_scalar])
    array_or_scalar = np.asarray(array_or_scalar)
    if array_or_scalar.ndim == 1:
        return array_or_scalar
    else:
        raise ValueError(
            f"Input must be a scalar or a 1D array, got {array_or_scalar.ndim} dimensions."
        )


def ensure_2d(array: np.ndarray) -> np.ndarray:
    """
    Ensure the input array has shape (num_pts, dim).

    Args:
        array (np.ndarray): Input array of shape (num_pts, dim) or (dim,).

    Returns:
        np.ndarray: A 2D array of shape (num_pts, dim).

    Raises:
        ValueError: If input is not a 1D or 2D array.
    """

    array = np.asarray(array)
    if array.ndim == 1:
        return array[np.newaxis, :]
    elif array.ndim == 2:
        return array
    else:
        raise ValueError(f"Input array must have 1 or 2 dimensions, got {array.ndim}.")



from .domain_mesh import DomainBilinearMesh2D, DomainMesh, DomainMesh2D
from .mesh_utils import bindex2dtuple, qindex2dtuple
from .subdomain_mesh import QuadMesh, SubdomainMesh2D

__all__ = [
    "SubdomainMesh2D",
    "QuadMesh",
    "bindex2dtuple",
    "qindex2dtuple",
    "DomainMesh",
    "DomainMesh2D",
    "DomainBilinearMesh2D",
]



from abc import ABC, abstractmethod
from typing import Any, List, Optional, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
import torchtt as tntt

from ttfemesh.basis.basis_utils import left_corner2index_ttmap, right_corner2index_ttmap
from ttfemesh.tt_tools.numeric import unit_vector_binary_tt
from ttfemesh.tt_tools.operations import zorder_kron
from ttfemesh.types import BoundarySide2D, TensorTrain


class Basis(ABC):
    """Abstract base class for basis functions."""

    @property
    @abstractmethod
    def index_range(self):  # pragma: no cover
        """Range of valid indices for the basis functions."""
        pass

    @abstractmethod
    def evaluate(self, *args, **kwargs) -> Any:  # pragma: no cover
        """Evaluate the basis function indexed with idx at a given point."""
        pass

    @abstractmethod
    def derivative(self, *args, **kwargs) -> Any:  # pragma: no cover
        """Evaluate the derivative of the basis function indexed with idx at a given point."""
        pass

    @abstractmethod
    def _validate(self, *args, **kwargs):  # pragma: no cover
        """Validate the basis function index."""
        pass

    @property
    @abstractmethod
    def dimension(self) -> int:  # pragma: no cover
        """The number of dimensions of the basis functions."""
        pass


class Basis1D(Basis):
    """Abstract base class for 1D basis functions on the reference element [-1, 1]."""

    def plot(self, idx: int, num_points: int = 100):
        """
        Plot the basis function indexed with idx.

        Args:
            idx (int): Index of the basis function.
            num_points (int): Number of points to plot.

        Returns:
            matplotlib.figure.Figure: The figure object containing the plot.
        """
        self._validate(idx)

        fig = plt.figure()
        x_vals = np.linspace(-1, 1, num_points)
        y_vals = np.array([self.evaluate(idx, x) for x in x_vals])
        plt.plot(x_vals, y_vals, label=f"Basis Function {idx}")

        plt.title("1D Basis Function on [-1, 1]")
        plt.xlabel("x")
        plt.ylabel("Basis Function Value")
        return fig

    @property
    def dimension(self) -> int:
        return 1

    @abstractmethod
    def get_element2global_ttmap(self, *args, **kwargs) -> TensorTrain:  # pragma: no cover
        """
        Get the TT-representation of a corner element index to global basis index map.

        Returns:
            TensorTrain: TT-representation of the corner to global index map.
        """
        pass

    @abstractmethod
    def get_all_element2global_ttmaps(
        self, *args, **kwargs
    ) -> Tuple[TensorTrain, ...]:  # pragma: no cover
        """
        Get the TT-representation for all corner elements in `index_range`
        to global basis index maps.

        Returns:
            Tuple[TensorTrain, ...]:  TT-representations of all corner-to-global index maps.
                First element is the map for index 0, second element is the map for index 1.
        """
        pass

    @abstractmethod
    def get_dirichlet_mask_left(self, *args, **kwargs) -> TensorTrain:  # pragma: no cover
        """
        Get the mask for the left Dirichlet boundary condition.

        Returns:
            TensorTrain: TT-representation of the Dirichlet mask.
        """
        pass

    @abstractmethod
    def get_dirichlet_mask_right(self, *args, **kwargs) -> TensorTrain:  # pragma: no cover
        """
        Get the mask for the right Dirichlet boundary condition.

        Returns:
            TensorTrain: TT-representation of the Dirichlet mask.
        """
        pass

    @abstractmethod
    def get_dirichlet_mask_left_right(self, *args, **kwargs) -> TensorTrain:  # pragma: no cover
        """
        Get the mask for the left and right Dirichlet boundary conditions.

        Returns:
            TensorTrain: TT-representation of the Dirichlet mask.
        """
        pass


class LinearBasis(Basis1D):
    """
    Linear basis functions on the reference element [-1, 1].
    The basis functions are defined as:
    - 0.5 * (1 - x) for idx = 0 (left basis function)
    - 0.5 * (1 + x) for idx = 1 (right basis function)

    Example:
        >>> from ttfemesh.basis import LinearBasis
        >>> basis1d = LinearBasis()
        >>> fig = basis1d.plot(0)
        >>> fig.show()
        >>> fig = basis1d.plot(1)
        >>> fig.show()
    """

    def evaluate(self, idx: int, x: float) -> float:
        """
        Evaluate the basis function at a given point.

        Args:
            idx (int): Index of the basis function.
            x (float): Point in [-1, 1] to evaluate the basis function.

        Returns:
            float: Value of the basis function at x.
        """
        self._validate(idx)

        if idx == 0:
            return 0.5 * (1 - x)

        return 0.5 * (1 + x)

    def derivative(self, idx: int, _: Optional[float] = None) -> float:
        """
        Evaluate the derivative of the basis function at a given point.
        Derivative is constant, so the point x is not used.

        Args:
            idx (int): Index of the basis function.
            _ (Optional[float]): Ignored.

        Returns:
            float: Derivative of the basis function at x.
        """
        self._validate(idx)
        return -0.5 if idx == 0 else 0.5

    @property
    def index_range(self):
        """
        Range of valid indices for the basis functions.
        0 for the left basis function, 1 for the right basis function.
        """
        return range(2)

    def get_element2global_ttmap(self, index: int, mesh_size_exponent: int) -> TensorTrain:
        """
        Get the TT-representation of a corner element index
        to global basis index map.

        Args:
            index (int): Index of the corner element (0 for left, 1 for right).
            mesh_size_exponent (int): Exponent of the 1D mesh size (length of TT).

        Returns:
            TensorTrain: TT-representation of the corner to global index map.

        Raises:
            ValueError: If the index is invalid.
        """
        self._validate(index)

        if index == 0:
            return left_corner2index_ttmap(mesh_size_exponent)
        elif index == 1:
            return right_corner2index_ttmap(mesh_size_exponent)

    def get_all_element2global_ttmaps(self, mesh_size_exponent: int) -> Tuple[TensorTrain, ...]:
        """
        Get TT-representations for all indices in `index_range`.

        Args:
            mesh_size_exponent (int): Exponent of the 1D mesh size (length of TT).

        Returns:
            Tuple[TensorTrain, ...]: TT-representations of all corner-to-global index maps.
                First element is the map for index 0, second element is the map for index 1.
        """
        return tuple(
            self.get_element2global_ttmap(index, mesh_size_exponent) for index in self.index_range
        )

    def get_dirichlet_mask_left(self, mesh_size_exponent: int) -> TensorTrain:
        """
        Get the mask for the left Dirichlet boundary condition.
        The mask is 1 everywhere except at the left boundary.

        Args:
            mesh_size_exponent (int): Exponent of the 1D mesh size (length of TT).

        Returns:
            TensorTrain: TT-representation of the Dirichlet mask.
        """

        mask = tntt.ones([2] * mesh_size_exponent) - unit_vector_binary_tt(mesh_size_exponent, 0)
        return mask

    def get_dirichlet_mask_right(self, mesh_size_exponent: int) -> TensorTrain:
        """
        Get the mask for the right Dirichlet boundary condition.
        The mask is 1 everywhere except at the right boundary.

        Args:
            mesh_size_exponent (int): Exponent of the 1D mesh size (length of TT).

        Returns:
            TensorTrain: TT-representation of the Dirichlet mask.
        """

        mask = tntt.ones([2] * mesh_size_exponent) - unit_vector_binary_tt(
            mesh_size_exponent, 2**mesh_size_exponent - 1
        )
        return mask

    def get_dirichlet_mask_left_right(self, mesh_size_exponent: int) -> TensorTrain:
        """
        Get the mask for the left and right Dirichlet boundary conditions.
        The mask is 1 everywhere except at the left and right boundaries.

        Args:
            mesh_size_exponent (int): Exponent of the 1D mesh size (length of TT).

        Returns:
            TensorTrain: TT-representation of the Dirichlet mask.
        """

        mask = (
            tntt.ones([2] * mesh_size_exponent)
            - unit_vector_binary_tt(mesh_size_exponent, 0)
            - unit_vector_binary_tt(mesh_size_exponent, 2**mesh_size_exponent - 1)
        )

        return mask

    def _validate(self, idx: int):
        if idx not in self.index_range:
            raise ValueError(
                f"Invalid basis function index: {idx}."
                f" Expected one of {list(self.index_range)}."
            )

    def __repr__(self):
        return "LinearBasis"


class TensorProductBasis(Basis):
    """
    Abstract base class for tensor product basis functions for arbitrary dimensions.
    Combines 1D basis functions to define basis functions in higher dimensions.
    """

    def __init__(self, basis_functions: List[Basis1D]):
        """
        Initialize the tensor product basis function.

        Args:
            basis_functions (List[BasisFunction1D]): List of 1D basis functions for each dimension.
        """
        self.basis_functions = basis_functions
        self._dimension = len(basis_functions)

    @property
    def dimension(self) -> int:
        return self._dimension

    @abstractmethod
    def get_element2global_ttmap(self, *args, **kwargs) -> TensorTrain:  # pragma: no cover
        """
        Get the TT-representation of a corner element index to global basis index map.

        Returns:
            TensorTrain: TT-representation of the corner to global index map.
        """
        pass

    @abstractmethod
    def get_all_element2global_ttmaps(self, *args, **kwargs) -> np.ndarray:  # pragma: no cover
        """
        Get the TT-representation for all corner elements in `index_range`
        to global basis index maps.

        Returns:
            np.ndarray: A 2D matrix of TT-representations, indexed by (i, j)
                where i and j are the indices of the basis functions in each dimension.
        """
        pass

    @abstractmethod
    def get_dirichlet_mask(self, *args, **kwargs) -> TensorTrain:  # pragma: no cover
        """
        Get the mask for the Dirichlet boundary condition on the specified sides.

        Returns:
            TensorTrain: TT-representation of the Dirichlet mask.
        """
        pass

    @property
    def index_range(self):
        return [bf.index_range for bf in self.basis_functions]

    def evaluate(self, idx: List[int], x: List[float]) -> float:
        """
        Evaluate the tensor product basis function at a given point.

        Args:
            idx (List[int]): Indices of the basis functions in each dimension.
            x (List[float]): Coordinates in the reference element [-1, 1]^d.

        Returns:
            float: Value of the tensor product basis function at x.
        """
        self._validate(idx)
        return np.prod([bf.evaluate(i, xi) for bf, i, xi in zip(self.basis_functions, idx, x)])

    def derivative(self, idx: List[int], x: List[float], dim: int) -> float:
        """
        Evaluate the partial derivative with respect to a given dimension.

        Args:
            idx (List[int]): Indices of the basis functions in each dimension.
            x (List[float]): Coordinates in the reference element [-1, 1]^d.
            dim (int): Dimension index (0-based) to differentiate.

        Returns:
            float: Value of the derivative at x.
        """
        self._validate(idx)
        if dim < 0 or dim >= self.dimension:
            raise ValueError(
                f"Invalid dimension index: {dim}, expected 0 <= dim < {self.dimension}"
            )

        result = 1.0
        for i, (bf, idx_xi, xi) in enumerate(zip(self.basis_functions, idx, x)):
            if i == dim:
                result *= bf.derivative(idx_xi, xi)
            else:
                result *= bf.evaluate(idx_xi, xi)
        return result

    def _validate(self, idx: List[int]):
        """Validate the basis function indices."""
        if len(idx) != self.dimension:
            raise ValueError(
                f"Invalid number of indices: expected {self.dimension}, got {len(idx)}"
            )
        for i, idx_i in enumerate(idx):
            self.basis_functions[i]._validate(idx_i)

    def __repr__(self):
        return f"TensorProductBasis(dim={self.dimension})"

    def plot(self, idx: List[int], num_points: int = 100):
        """
        Plot the tensor product basis function indexed with idx.

        Args:
            idx (List[int]): Indices of the basis functions in each dimension.
            num_points (int): Number of points to plot.

        Raises:
            NotImplementedError: If the dimension is not 2.
        """
        self._validate(idx)

        if self.dimension == 2:
            return self._plot2d(idx, num_points)
        else:
            raise NotImplementedError(
                "Tensor product plotting" " is only supported for 2D basis functions."
            )

    def _plot2d(self, idx: List[int], num_points: int = 100):
        """
        Plot the tensor product basis function indexed with idx in 2D.

        Args:
            idx (List[int]): Indices of the basis functions in each dimension.
            num_points (int): Number of points to plot.

        Returns:
            matplotlib.figure.Figure: The figure object containing the plot.
        """
        fig = plt.figure()
        x_vals = np.linspace(-1, 1, num_points)
        y_vals = np.linspace(-1, 1, num_points)
        X, Y = np.meshgrid(x_vals, y_vals)
        Z = np.zeros_like(X)
        for i in range(num_points):
            for j in range(num_points):
                Z[i, j] = self.evaluate(idx, [X[i, j], Y[i, j]])
        plt.contourf(X, Y, Z, levels=20)
        plt.colorbar()
        plt.title("2D Tensor Product Basis Function")
        plt.xlabel("x")
        plt.ylabel("y")
        return fig


class BilinearBasis(TensorProductBasis):
    """
    Bilinear basis functions on the reference element [-1, 1]^2.
    The basis functions are defined as:
    - 0.25 * (1 - x) * (1 - y) for idx = (0, 0)
    - 0.25 * (1 + x) * (1 - y) for idx = (1, 0)
    - 0.25 * (1 - x) * (1 + y) for idx = (0, 1)
    - 0.25 * (1 + x) * (1 + y) for idx = (1, 1)

    Example:
        >>> from ttfemesh.basis import BilinearBasis
        >>> basis2d = BilinearBasis()
        >>> fig = basis2d.plot([0, 0])
        >>> fig.show()
        >>> fig = basis2d.plot([0, 1])
        >>> fig.show()
        >>> fig = basis2d.plot([1, 0])
        >>> fig.show()
        >>> fig = basis2d.plot([1, 1])
        >>> fig.show()
    """

    def __init__(self):
        super().__init__([LinearBasis(), LinearBasis()])

    def get_element2global_ttmap(self, index: List[int], mesh_size_exponent: int) -> TensorTrain:
        """
        Get the TT-representation of a corner element index to global basis index map.

        Args:
            index (List[int]): Indices of the corner element
                ((0, 0) for lower left,
                (1, 0) for lower right,
                (0, 1) for upper left,
                (1, 1) for upper right).
            mesh_size_exponent (int): Exponent of the 1D mesh size.

        Returns:
            TensorTrain: TT-representation of the corner to global index map.

        Raises:
            ValueError: If the index is invalid.
        """

        self._validate(index)
        return zorder_kron(
            self.basis_functions[0].get_element2global_ttmap(index[0], mesh_size_exponent),
            self.basis_functions[1].get_element2global_ttmap(index[1], mesh_size_exponent),
        )

    def get_all_element2global_ttmaps(self, mesh_size_exponent: int) -> np.ndarray:
        """
        Get the TT-representation for all corner elements in `index_range`
        to global basis index maps.

        Args:
            mesh_size_exponent (int): Exponent of the 1D mesh size.

        Returns:
            np.ndarray: A 2D matrix of TT-representations, indexed by (i, j)
                where i and j are the indices of the basis functions in each dimension.
        """
        return np.array(
            [
                [
                    zorder_kron(
                        self.basis_functions[0].get_element2global_ttmap(i, mesh_size_exponent),
                        self.basis_functions[1].get_element2global_ttmap(j, mesh_size_exponent),
                    )
                    for j in self.index_range[1]
                ]
                for i in self.index_range[0]
            ]
        )

    def get_dirichlet_mask(
        self, mesh_size_exponent: int, *sides: Union[BoundarySide2D, int]
    ) -> TensorTrain:
        """
        Get the mask for the Dirichlet 2D boundary condition on the specified sides.

        Note that the sides are considered to be ordered as follows:
        bottom (side 0), right (side 1), top (side 2), left (side 3).
        This is important for the boundary condition to work correctly.
        It may lead to confusion if, e.g., your side 0 is visually
        the right edge of the domain.

        Args:
            mesh_size_exponent (int): Exponent of the 1D mesh size.
            *sides (Union[BoundarySide2D, int]): Boundary sides to apply the Dirichlet condition.
                Specify either as integers (0 for bottom, 1 for right, 2 for top, 3 for left)
                or as BoundarySide2D enums.

        Returns:
            TensorTrain: TT-representation of the Dirichlet mask.

        Raises:
            ValueError: If no sides are specified or if an invalid side is given.
        """

        if not sides:
            raise ValueError("At least one boundary side must be specified.")

        sides_ = [BoundarySide2D(side) if isinstance(side, int) else side for side in sides]
        xmask = tntt.ones([2] * mesh_size_exponent)
        if BoundarySide2D.LEFT in sides_ and BoundarySide2D.RIGHT in sides_:
            xmask = self.basis_functions[0].get_dirichlet_mask_left_right(mesh_size_exponent)
        elif BoundarySide2D.LEFT in sides_:
            xmask = self.basis_functions[0].get_dirichlet_mask_left(mesh_size_exponent)
        elif BoundarySide2D.RIGHT in sides_:
            xmask = self.basis_functions[0].get_dirichlet_mask_right(mesh_size_exponent)

        ymask = tntt.ones([2] * mesh_size_exponent)
        if BoundarySide2D.BOTTOM in sides_ and BoundarySide2D.TOP in sides_:
            ymask = self.basis_functions[1].get_dirichlet_mask_left_right(mesh_size_exponent)
        elif BoundarySide2D.BOTTOM in sides_:
            ymask = self.basis_functions[1].get_dirichlet_mask_left(mesh_size_exponent)
        elif BoundarySide2D.TOP in sides_:
            ymask = self.basis_functions[1].get_dirichlet_mask_right(mesh_size_exponent)

        return zorder_kron(xmask, ymask)

    def __repr__(self):
        return super().__repr__() + "::" + "BilinearBasis"



from typing import List

import torch

from ttfemesh.types import TensorTrain


def integer_to_little_endian(
    length: int,
    num: int,
) -> List[int]:
    """
    Convert an integer to its little-endian binary representation.

    Args:
        length (int): Number of bits to represent the integer.
        num (int): An integer.

    Returns:
        List[int]: A list of bits representing the integer in little-endian format.

    Raises:
        ValueError: If num is negative or greater than or equal to 2^length.
    """

    if num < 0:
        raise ValueError(f"Input number ({num}) must be non-negative.")
    if num >= 2**length:
        raise ValueError(f"Input number {num} must be less than 2^{length}.")

    return [(num >> i) & 1 for i in range(length)]


def unit_vector_binary_tt(length: int, index: int) -> TensorTrain:
    """
    Generate a unit vector with a 1 at the specified index for a binary quantized tensor train.

    Args:
        length (int): Length of the tensor train.
        index (int): Index of the position of the 1 in the unit vector.

    Returns:
        TensorTrain: A tensor train of dimension length representing the unit vector.

    Raises:
        ValueError: If index is negative or greater than or equal to 2^length.
    """

    if index < 0:
        raise ValueError(f"Index ({index}) must be non-negative.")
    if index >= 2**length:
        raise ValueError(f"Index ({index}) must be less than 2^{length}.")

    cores = []
    binidx = integer_to_little_endian(length, index)
    for j in range(length):
        core = torch.zeros(1, 2, 1)
        core[0, binidx[j], 0] = 1.0
        cores.append(core)

    return TensorTrain(cores)



import matplotlib.pyplot as plt
import numpy as np


def plot_curve_with_tangents(curve, title, num_points=50):
    """
    Plot a curve and its tangent vectors on a 2D plane.

    Args:
        curve (Curve): An instance of a Curve object.
        title (str): Title of the plot.
        num_points (int): Number of points to sample along the curve.
    """
    t_vals = np.linspace(-1, 1, num_points)
    points = curve.evaluate(t_vals)
    tangents = curve.tangent(t_vals)

    plt.plot(points[:, 0], points[:, 1], label="Curve", color="blue")

    for i in range(0, num_points, max(1, num_points // 10)):
        p = points[i]
        t = tangents[i]
        plt.quiver(
            p[0],
            p[1],
            t[0],
            t[1],
            angles="xy",
            scale_units="xy",
            scale=1.0,
            color="red",
            width=0.003,
            label="Tangent" if i == 0 else "",
        )

    plt.title(title)
    plt.xlabel("x")
    plt.ylabel("y")
    plt.axis("equal")
    plt.legend()
    plt.grid()



from typing import Callable

import numpy as np

from ttfemesh.tt_tools.meshgrid import range_meshgrid2d
from ttfemesh.tt_tools.operations import zorder_linfunc2d
from ttfemesh.types import TensorTrain


def interpolate_linear2d(
    func: Callable[[np.ndarray], float], mesh_size_exponent: int
) -> TensorTrain:
    """
    Interpolate a function on a 2D grid using linear interpolation.
    The function takes a quaternary argument index, corresponding to an index on the grid,
    arranged in the z-order, and returns a float value,
    corresponding to the function value at that point.

    Args:
        func (Callable[[ndarray], float]): Function to interpolate.
        mesh_size_exponent (int): Exponent of the 1D grid size.

    Returns:
        TensorTrain resulting from the linear interpolation of the function.
    """

    num_total = 2**mesh_size_exponent
    index0 = np.zeros(mesh_size_exponent, dtype=int)
    indexn0 = np.zeros(mesh_size_exponent, dtype=int)
    index0n = np.zeros(mesh_size_exponent, dtype=int)
    indexn0[::] = 1
    index0n[::] = 2

    c = func(index0)
    cx = (func(indexn0) - c) / (num_total - 1.0)
    cy = (func(index0n) - c) / (num_total - 1.0)

    XX, YY = range_meshgrid2d(mesh_size_exponent)
    result = zorder_linfunc2d(c, cx, XX, cy, YY)

    return result



from abc import ABC, abstractmethod
from typing import Tuple

from ttfemesh.domain.curve import Line2D
from ttfemesh.domain.subdomain import Quad, Subdomain2D


class SubdomainFactory(ABC):
    """
    Abstract base class for a subdomain factory.
    """

    @staticmethod  # noqa
    @abstractmethod
    def create(*args, **kwargs) -> Subdomain2D:  # pragma: no cover
        pass


class RectangleFactory(SubdomainFactory):
    """
    Factory class for creating rectangle subdomains.

    Example:
        >>> from ttfemesh.domain import RectangleFactory
        >>> lower_left = (0, 0)
        >>> upper_right = (2, 1)
        >>> rectangle = RectangleFactory.create(lower_left, upper_right)
        >>> rectangle.plot()
    """

    @staticmethod  # noqa
    def create(bottom_left: Tuple[float, float], top_right: Tuple[float, float]) -> Quad:
        """
        Create a rectangle subdomain defined by the bottom-left and top-right corners.

        Args:
            bottom_left (Tuple[float, float]): Coordinates of the bottom-left corner.
            top_right (Tuple[float, float]): Coordinates of the top-right corner.
        Returns:
            Quad: A rectangle subdomain.
        """

        x0, y0 = bottom_left
        x1, y1 = top_right
        points = [(x0, y0), (x1, y0), (x1, y1), (x0, y1)]
        return Quad([Line2D(points[i], points[(i + 1) % 4]) for i in range(4)])


class QuadFactory(SubdomainFactory):
    """
    Factory class for creating quadrilateral subdomains.
    Note that the lines are considered to be ordered as follows:
    bottom (line 0), right (line 1), top (line 2), left (line 3).
    This is important for the boundary condition to work correctly.
    It may lead to confusion if, e.g., your line 0 is visually
    the right edge of the domain.

    Example:
        >>> from ttfemesh.domain import QuadFactory
        >>> p1 = (3, 0)
        >>> p2 = (1, -3)
        >>> p3 = (7, -4)
        >>> p4 = (5, -1)
        >>> quad3 = QuadFactory.create(p1, p2, p3, p4)
        >>> quad3.plot()
    """

    @staticmethod  # noqa
    def create(
        p1: Tuple[float, float],
        p2: Tuple[float, float],
        p3: Tuple[float, float],
        p4: Tuple[float, float],
    ) -> Quad:
        """
        Create a trapezoid subdomain defined by the four corner points.
        Points must be ordered counter-clockwise.

        Args:
            p1 (Tuple[float, float]): Coordinates of the first corner.
            p2 (Tuple[float, float]): Coordinates of the second corner.
            p3 (Tuple[float, float]): Coordinates of the third corner.
            p4 (Tuple[float, float]): Coordinates of the fourth corner.
        Returns:
            Quad: A trapezoid subdomain.
        """

        points = (p1, p2, p3, p4)
        return Quad([Line2D(points[i], points[(i + 1) % 4]) for i in range(4)])



from abc import ABC, abstractmethod
from typing import Optional, Sequence

import matplotlib.pyplot as plt
import numpy as np

from ttfemesh.domain.boundary_condition import BoundaryCondition, DirichletBoundary2D
from ttfemesh.domain.subdomain import Subdomain, Subdomain2D
from ttfemesh.domain.subdomain_connection import (
    CurveConnection2D,
    SubdomainConnection,
    SubdomainConnection2D,
    VertexConnection2D,
)


class Domain(ABC):
    """Domain class that contains subdomains and their connections."""

    def __init__(
        self,
        subdomains: Sequence[Subdomain],
        connections: Sequence[SubdomainConnection],
        boundary_condition: Optional[BoundaryCondition] = None,
    ):
        """
        Initialize a domain with subdomains and their connections.

        Args:
            subdomains (Sequence[Subdomain]): List of subdomains in the domain.
            connections (Sequence[SubdomainConnection]): List of connections between subdomains.
            boundary_condition (Optional[BoundaryCondition]): Optional boundary condition.
        """
        self.subdomains = subdomains
        self.connections = connections
        self.boundary_condition = boundary_condition
        self.validate()

    @property
    def num_subdomains(self) -> int:
        """Number of subdomains in the domain."""
        return len(self.subdomains)

    @property  # noqa
    def num_connections(self) -> int:
        """Number of connections in the domain."""
        return len(self.connections)

    def get_connections(self) -> Sequence[SubdomainConnection]:
        """Get the list of connections."""
        return self.connections

    def get_subdomain(self, idx) -> Subdomain:
        """
        Get a subdomain by index.

        Args:
            idx (int): Index of the subdomain to get.

        Returns:
            Subdomain: The subdomain at the given index.

        Raises:
            ValueError: If the index is out of bounds.
        """
        if idx < 0 or idx >= len(self.subdomains):
            raise ValueError(
                f"Invalid subdomain index: {idx}. " f"Must be in [0, {len(self.subdomains)})."
            )

        return self.subdomains[idx]

    def validate(self):
        """Validates that the connections are consistent with the subdomains."""
        for connection in self.connections:
            connection.validate(self.subdomains)
        if self.boundary_condition:
            self.boundary_condition.validate(self.subdomains)

    def __repr__(self):
        return f"Domain({len(self.subdomains)} subdomains, {len(self.connections)} connections)"

    def plot(self):
        "Plot the subdomains with connections."
        for subdomain in self.subdomains:
            subdomain.plot()

    @property
    @abstractmethod
    def dimension(self):  # pragma: no cover
        """The dimension of the domain."""
        pass


class Domain2D(Domain):
    """A 2D domain with subdomains and their connections.

    Example:
    >>> from ttfemesh.domain import RectangleFactory, CurveConnection2D, VertexConnection2D
    >>> from ttfemesh.domain import DirichletBoundary2D, Domain2D
    >>> lower_left = (0, 0)
    >>> upper_right = (2, 1)
    >>> rectangle1 = RectangleFactory.create(lower_left, upper_right)

    >>> lower_left = (2, 0)
    >>> upper_right = (3, 1)
    >>> rectangle2 = RectangleFactory.create(lower_left, upper_right)

    >>> lower_left = (-2, 1)
    >>> upper_right = (0, 2)
    >>> rectangle3 = RectangleFactory.create(lower_left, upper_right)

    >>> domain_idxs = [0, 1]
    >>> curve_idxs = [1, 3]
    >>> edge = CurveConnection2D(domain_idxs, curve_idxs)

    >>> vertex_idxs = [(0, 3, "start"), (2, 0, "end")]
    >>> vertex = VertexConnection2D(vertex_idxs)

    >>> bc = DirichletBoundary2D([(1, 1), (2, 3)])

    >>> domain = Domain2D([rectangle1, rectangle2, rectangle3], [edge, vertex], bc)
    >>> domain.plot()
    """

    def __init__(
        self,
        subdomains: Sequence[Subdomain2D],
        connections: Sequence[SubdomainConnection2D],
        boundary_condition: Optional[DirichletBoundary2D] = None,
    ):
        """
        Initialize a 2D domain with subdomains and their connections.

        Args:
            subdomains (Sequence[Subdomain2D]): List of 2D subdomains in the domain.
            connections (Sequence[SubdomainConnection2D]): List of connections between subdomains.
            boundary_condition (Optional[DirichletBoundary2D]): Optional 2D boundary condition.
        """
        super().__init__(subdomains, connections, boundary_condition)

    @property
    def dimension(self) -> int:
        """The dimension of the domain."""
        return 2

    def plot(self, num_points=100):
        """
        Plot the domain and its subdomains with connections.

        Args:
            num_points (int): Number of points to sample.
        """

        for subdomain in self.subdomains:
            for curve in subdomain.curves:
                t = np.linspace(-1, 1, 100)
                points = np.array(curve.evaluate(t))
                plt.plot(points[:, 0], points[:, 1], label="Subdomain")

        for connection in self.connections:
            if isinstance(connection, VertexConnection2D):
                shared_vertex = connection.get_shared_vertex(self.subdomains)
                plt.plot(shared_vertex[0], shared_vertex[1], "ro", label="Shared Vertex")
            elif isinstance(connection, CurveConnection2D):
                curve = connection.get_shared_curve(self.subdomains)
                curve_points = curve.evaluate(np.linspace(-1, 1, num_points))
                plt.plot(curve_points[:, 0], curve_points[:, 1], "g--", label="Shared Curve")

        if self.boundary_condition:
            for subdomain_idx, curve_idx in self.boundary_condition.boundary:
                curve = self.subdomains[subdomain_idx].curves[curve_idx]
                curve_points = curve.evaluate(np.linspace(-1, 1, num_points))
                plt.plot(curve_points[:, 0], curve_points[:, 1], "k--", label="Boundary Condition")

        plt.axis("equal")
        plt.title("Domain Plot")
        plt.show()

    def __repr__(self):
        return f"Domain2D({len(self.subdomains)} subdomains, {len(self.connections)} connections)"



import warnings
from abc import ABC, abstractmethod
from typing import Callable, Optional, Tuple

import matplotlib.pyplot as plt
import numpy as np
import torch

from ttfemesh.domain import Quad, Subdomain, Subdomain2D
from ttfemesh.mesh.mesh_utils import qindex2dtuple as index_map2d
from ttfemesh.quadrature.quadrature import QuadratureRule, QuadratureRule2D
from ttfemesh.tt_tools import interpolate_linear2d
from ttfemesh.tt_tools.tensor_cross import (
    TTCrossConfig,
    anova_init_tensor_train,
    gen_teneva_indices,
    tensor_train_cross_approximation,
)
from ttfemesh.types import TensorTrain
from ttfemesh.utils.array import ensure_2d


class SubdomainMesh(ABC):
    """Subdomain mesh for a finite element problem."""

    def __init__(
        self,
        subdomain: Subdomain,
        quadrature_rule: QuadratureRule,
        mesh_size_exponent: int,
        tt_cross_config: Optional[TTCrossConfig] = None,
    ):
        """
        Initialize a subdomain mesh.

        Args:
            subdomain (Subdomain): The subdomain to mesh.
            quadrature_rule (QuadratureRule): The quadrature rule to use.
            mesh_size_exponent (int): The exponent of the discretization size.
                The discretization size is 2**(mesh_size_exponent) per dimension.
            tt_cross_config (Optional[TTCrossConfig]):
                Configuration for the tensor train cross approximation.
                If None, default configuration is used.
        """
        self.subdomain = subdomain
        self.quadrature_rule = quadrature_rule
        self.mesh_size_exponent = mesh_size_exponent
        self._index_map: Optional[Callable] = None

        self._tt_cross_config = tt_cross_config
        if self._tt_cross_config is None:
            self._tt_cross_config = TTCrossConfig(info={})

    @abstractmethod
    def ref2domain_map(self, xi) -> np.ndarray:  # pragma: no cover
        """
        Return the reference to domain map.

        Args:
            xi: The reference coordinates in the element.

        Returns:
            np.ndarray: The physical coordinates in the domain.
        """
        pass

    @abstractmethod
    def ref2element_map(self, index, xi) -> np.ndarray:  # pragma: no cover
        """
        Return the element transformation function.

        Args:
            index: The index of the element.
            xi: The reference coordinates in the element.

        Returns:
            np.ndarray: The physical coordinates in the element.
        """
        pass

    @abstractmethod
    def ref2domain_jacobian(self, xi) -> np.ndarray:  # pragma: no cover
        """
        Return the Jacobian function for the domain transformation.

        Args:
            xi: The reference coordinates in the element.

        Returns:
            np.ndarray: The Jacobian of the domain transformation.
        """
        pass

    @abstractmethod
    def get_jacobian_tensor_trains(self):  # noqa # pragma: no cover
        """
        Compute the tensor network approximating the Jacobian evaluated on all elements.
        The tensor index corresponds to the element index.
        This is done for each Jacobian component and each quadrature point within the element.
        The output is thus a total of 4*(num_quadrature_points_per_element) tensor networks.
        """
        pass

    @abstractmethod
    def plot(self):  # pragma: no cover
        """Plot the subdomain mesh."""
        pass

    @abstractmethod
    def _validate_idxs(self, *indices):  # noqa # pragma: no cover
        """Validate indices."""
        pass

    @abstractmethod
    def _validate_ref_coords(self, *coords, **kwargs):  # noqa # pragma: no cover
        """Validate reference element coordinates."""
        pass

    @property
    @abstractmethod
    def dimension(self):  # pragma: no cover
        """Return the dimension of the mesh."""
        pass


class SubdomainMesh2D(SubdomainMesh):
    """Subdomain mesh for a 2D finite element problem.

    Example:
    >>> from ttfemesh.domain import RectangleFactory
    >>> from ttfemesh.quadrature import GaussLegendre2D
    >>> from ttfemesh.mesh import SubdomainMesh2D

    >>> lower_left = (0, 0)
    >>> upper_right = (2, 1)
    >>> rectangle = RectangleFactory.create(lower_left, upper_right)
    >>> quadrature_rule = GaussLegendre2D()
    >>> mesh_size_exponent = 3
    >>> mesh = SubdomainMesh2D(rectangle, quadrature_rule, mesh_size_exponent)
    >>> mesh.plot()
    """

    def __init__(
        self,
        subdomain: Subdomain2D,
        quadrature_rule: QuadratureRule2D,
        mesh_size_exponent: int,
        tt_cross_config: Optional[TTCrossConfig] = None,
    ):
        """
        Initialize a 2D subdomain mesh.

        Args:
            subdomain (Subdomain2D): The 2D subdomain to mesh.
            quadrature_rule (QuadratureRule2D): The quadrature rule to use.
            mesh_size_exponent (int): The exponent of the discretization size.
                The discretization size is 2**(mesh_size_exponent) per dimension.
            tt_cross_config (Optional[TTCrossConfig]):
                Configuration for the tensor train cross approximation.
                If None, default configuration is used.

        """
        super().__init__(subdomain, quadrature_rule, mesh_size_exponent, tt_cross_config)

        self._num_points1d = 2**mesh_size_exponent
        self._grid_step1d = 2.0 / (self._num_points1d - 1)
        self._index_map = index_map2d
        self._tca_strategy = self.__tca_default

    @property
    def dimension(self):
        """Return the dimension of the mesh."""
        return 2

    @property
    def num_points1d(self):
        """Number of points per dimension."""
        return self._num_points1d

    @property
    def num_points(self):
        """Total number of points."""
        return self.num_points1d**2

    @property
    def num_elements1d(self):
        """Number of elements per dimension."""
        return self.num_points1d - 1

    @property
    def num_elements(self):
        """Total number of elements."""
        return (self.num_points1d - 1) ** 2

    @property
    def index_map(self):
        """Index map for binary index to 2D tuple."""
        return self._index_map

    @property
    def tt_cross_config(self):
        """Configuration for the tensor train cross approximation."""
        return self._tt_cross_config

    @tt_cross_config.setter
    def tt_cross_config(self, config: TTCrossConfig):
        """Set the configuration for the tensor train cross approximation."""
        self._tt_cross_config = config

    def ref2domain_map(self, xi_eta: np.ndarray) -> np.ndarray:
        """
        Map from reference quadrilateral [-1, 1]^2 to domain.

        Args:
            xi_eta (np.ndarray): The reference coordinates in the quadrilateral.
                Of shape (num_points, 2) or (2,).

        Returns:
            np.ndarray: The physical coordinates in the domain.
                Of shape (num_points, 2).
        """

        xi_eta_ = ensure_2d(xi_eta)
        self._validate_ref_coords(xi_eta_)
        xi, eta = xi_eta_[:, 0], xi_eta_[:, 1]

        assert isinstance(self.subdomain, Subdomain2D)  # nosec

        side0 = self.subdomain.curves[0]
        side1 = self.subdomain.curves[1]
        side2 = self.subdomain.curves[2]
        side3 = self.subdomain.curves[3]

        side0_vals = side0(xi)
        side1_vals = side1(eta)
        side2_vals = side2(-xi)
        side3_vals = side3(-eta)

        side0_x, side0_y = side0_vals[:, 0], side0_vals[:, 1]
        side1_x, side1_y = side1_vals[:, 0], side1_vals[:, 1]
        side2_x, side2_y = side2_vals[:, 0], side2_vals[:, 1]
        side3_x, side3_y = side3_vals[:, 0], side3_vals[:, 1]

        side0_start = side0.get_start()
        side1_start = side1.get_start()
        side2_start = side2.get_start()
        side3_start = side3.get_start()

        side0_x_start, side0_y_start = side0_start[0], side0_start[1]
        side1_x_start, side1_y_start = side1_start[0], side1_start[1]
        side2_x_start, side2_y_start = side2_start[0], side2_start[1]
        side3_x_start, side3_y_start = side3_start[0], side3_start[1]

        N_xi_eta_x = (
            0.5 * (1.0 - eta) * side0_x
            + 0.5 * (1.0 + xi) * side1_x
            + 0.5 * (1.0 + eta) * side2_x
            + 0.5 * (1.0 - xi) * side3_x
            - 0.25 * (1.0 - xi) * (1.0 - eta) * side0_x_start
            - 0.25 * (1.0 + xi) * (1.0 - eta) * side1_x_start
            - 0.25 * (1.0 + xi) * (1.0 + eta) * side2_x_start
            - 0.25 * (1.0 - xi) * (1.0 + eta) * side3_x_start
        )

        N_xi_eta_y = (
            0.5 * (1.0 - eta) * side0_y
            + 0.5 * (1.0 + xi) * side1_y
            + 0.5 * (1.0 + eta) * side2_y
            + 0.5 * (1.0 - xi) * side3_y
            - 0.25 * (1.0 - xi) * (1.0 - eta) * side0_y_start
            - 0.25 * (1.0 + xi) * (1.0 - eta) * side1_y_start
            - 0.25 * (1.0 + xi) * (1.0 + eta) * side2_y_start
            - 0.25 * (1.0 - xi) * (1.0 + eta) * side3_y_start
        )

        N_xi_eta = np.stack([N_xi_eta_x, N_xi_eta_y], axis=-1)

        return N_xi_eta

    def ref2element_map(self, index: Tuple[int, int], xi_eta: np.ndarray) -> np.ndarray:
        """
        Map from reference quadrilateral [-1, 1]^2 to element indexed by (index_x, index_y).

        Args:
            index (Tuple[int, int]): The 2D index of the element.
            xi_eta (np.ndarray): The reference coordinates in the quadrilateral.
                Of shape (num_points, 2) or (2,).

        Returns:
            np.ndarray: The physical coordinates in the element.
                Of shape (num_points, 2).
        """

        xi_eta_ = ensure_2d(xi_eta)
        self._validate_idxs(*index)
        self._validate_ref_coords(xi_eta_)

        index_x, index_y = index
        xi, eta = xi_eta_[:, 0], xi_eta_[:, 1]
        offset_xi = -1.0 + index_x * self._grid_step1d
        offset_eta = -1.0 + index_y * self._grid_step1d
        xi_rescaled = offset_xi + 0.5 * (1.0 + xi) * self._grid_step1d
        eta_rescaled = offset_eta + 0.5 * (1.0 + eta) * self._grid_step1d

        return self.ref2domain_map(np.column_stack((xi_rescaled, eta_rescaled)))

    def ref2domain_jacobian(self, xi_eta: np.ndarray) -> np.ndarray:
        """
        Compute the Jacobian of the reference to domain map.

        Args:
            xi_eta (np.ndarray): The reference coordinates in the quadrilateral.
                Of shape (num_points, 2) or (2,).

        Returns:
            np.ndarray: The Jacobian of the reference to domain map.
                Of shape (num_points, 2, 2).
        """

        xi_eta_ = ensure_2d(xi_eta)
        self._validate_ref_coords(xi_eta_)

        xi, eta = xi_eta_[:, 0], xi_eta_[:, 1]

        assert isinstance(self.subdomain, Subdomain2D)  # nosec

        side0 = self.subdomain.curves[0]
        side1 = self.subdomain.curves[1]
        side2 = self.subdomain.curves[2]
        side3 = self.subdomain.curves[3]

        side0_vals = side0(xi)
        side1_vals = side1(eta)
        side2_vals = side2(-xi)
        side3_vals = side3(-eta)

        side0_tangent = side0.tangent(xi)
        side1_tangent = side1.tangent(eta)
        side2_tangent = -side2.tangent(-xi)
        side3_tangent = -side3.tangent(-eta)

        dxi_N = (
            0.5 * (1.0 - eta)[:, None] * side0_tangent
            + 0.5 * side1_vals
            + 0.5 * (1.0 + eta)[:, None] * side2_tangent
            - 0.5 * side3_vals
            + 0.25 * (1.0 - eta)[:, None] * side0.get_start()
            - 0.25 * (1.0 - eta)[:, None] * side1.get_start()
            - 0.25 * (1.0 + eta)[:, None] * side2.get_start()
            + 0.25 * (1.0 + eta)[:, None] * side3.get_start()
        )

        deta_N = (
            -0.5 * side0_vals
            + 0.5 * (1.0 + xi)[:, None] * side1_tangent
            + 0.5 * side2_vals
            + 0.5 * (1.0 - xi)[:, None] * side3_tangent
            + 0.25 * (1.0 - xi)[:, None] * side0.get_start()
            + 0.25 * (1.0 + xi)[:, None] * side1.get_start()
            - 0.25 * (1.0 + xi)[:, None] * side2.get_start()
            - 0.25 * (1.0 - xi)[:, None] * side3.get_start()
        )

        jacobian = np.stack([dxi_N[:, 0], deta_N[:, 0], dxi_N[:, 1], deta_N[:, 1]], axis=-1)
        jacobian = jacobian.reshape(-1, 2, 2)

        return jacobian

    def ref2element_jacobian(self, index: Tuple[int, int], xi_eta: np.ndarray) -> np.ndarray:
        """
        Compute the Jacobian of the reference to element map.

        Args:
            index (Tuple[int, int]): The 2D index of the element.
            xi_eta (np.ndarray): The reference coordinates in the quadrilateral.
                Of shape (num_points, 2) or (2,).

        Returns:
            np.ndarray: The Jacobian of the reference to element map.
                Of shape (num_points, 2, 2).
        """

        xi_eta_ = ensure_2d(xi_eta)
        self._validate_idxs(*index)
        self._validate_ref_coords(xi_eta_)

        index_x, index_y = index
        xi, eta = xi_eta_[:, 0], xi_eta_[:, 1]
        offset_xi = -1.0 + index_x * self._grid_step1d
        offset_eta = -1.0 + index_y * self._grid_step1d
        xi_rescaled = offset_xi + 0.5 * (1.0 + xi) * self._grid_step1d
        eta_rescaled = offset_eta + 0.5 * (1.0 + eta) * self._grid_step1d

        jacobian_domain = self.ref2domain_jacobian(np.column_stack((xi_rescaled, eta_rescaled)))
        scaling = 0.5 * self._grid_step1d
        jacobian_rescaled = jacobian_domain * scaling

        return jacobian_rescaled

    def get_jacobian_tensor_trains(self) -> np.ndarray:  # noqa
        """
        Compute the tensor train approximating the Jacobian evaluated on all elements.
        The tensor index corresponds to the element index.
        This is done for each Jacobian component and each quadrature point within the element.
        The output is thus a total of 4*(num_quadrature_points_per_element) tensor trains.

        Returns:
            np.ndarray: 3D array of tensor trains for the Jacobian components.
               Indexing: [quadrature_point_index, component_index_i, component_index_j].
        """

        quadrature_points, _ = self.quadrature_rule.get_points_weights()

        jacobian_tensor_trains: np.ndarray = np.ndarray(
            (len(quadrature_points), 2, 2), dtype=object
        )
        for q, quad_point in enumerate(quadrature_points):
            for i in range(2):
                for j in range(2):
                    eval_point = np.array([quad_point])

                    def cross_func(idx):
                        return self._cross_func(idx, eval_point)[:, i, j]

                    # out of bounds intentional
                    with warnings.catch_warnings():
                        warnings.simplefilter("ignore", UserWarning)
                        jac_ij = self._tca_strategy(cross_func)

                    jacobian_tensor_trains[q, i, j] = jac_ij
        return jacobian_tensor_trains

    def get_jacobian_det_tensor_trains(self) -> np.ndarray:  # noqa
        """
        Compute the tensor train approximating
        the Jacobian determinants evaluated on all elements.

        Returns:
            np.ndarray: 1D array of tensor trains for the Jacobian determinants.
                Indexing: [quadrature_point_index].
        """
        quadrature_points, _ = self.quadrature_rule.get_points_weights()

        det_tensor_trains: np.ndarray = np.ndarray((len(quadrature_points)), dtype=object)
        for q, quad_point in enumerate(quadrature_points):
            eval_point = np.array([quad_point])

            def cross_func(idx):
                return np.linalg.det(self._cross_func(idx, eval_point))

            # out of bounds intentional
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", UserWarning)
                jac_det = self._tca_strategy(cross_func)

            det_tensor_trains[q] = jac_det

        return det_tensor_trains

    def get_jacobian_invdet_tensor_trains(self) -> np.ndarray:  # noqa
        """
        Compute the tensor train approximating
        the inverse of Jacobian determinants evaluated on all elements.

        Returns:
            np.ndarray: 1D array of tensor trains for the inverse Jacobian determinants.
                Indexing: [quadrature_point_index].
        """
        quadrature_points, _ = self.quadrature_rule.get_points_weights()

        invdet_tensor_trains = []
        for quad_point in quadrature_points:
            eval_point = np.array([quad_point])

            def cross_func(idx):
                return 1.0 / np.linalg.det(self._cross_func(idx, eval_point))

            # out of bounds intentional
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", UserWarning)
                jac_invdet = self.__tca_default(cross_func)  # linear interpolation never works here

            invdet_tensor_trains.append(jac_invdet)

        return np.array(invdet_tensor_trains)

    def get_jacobian_tensors(self) -> np.ndarray:
        """
        Compute the Jacobians evaluated on all elements and all quadrature points.

        Returns:
            np.ndarray: Jacobians.
                Of shape (num_elements_x+1, num_elements_y+1, num_quadrature_points, 2, 2).

        Warning:
            This method is not efficient for large meshes.
            Intended only for small meshes for testing.
        """

        quadrature_points, _ = self.quadrature_rule.get_points_weights()
        num_elements_x = self.num_elements1d
        num_elements_y = self.num_elements1d
        num_quadrature_points = quadrature_points.shape[0]
        jacobians = np.empty((num_elements_x + 1, num_elements_y + 1, num_quadrature_points, 2, 2))

        for index_x in range(num_elements_x + 1):
            for index_y in range(num_elements_y + 1):
                index = (index_x, index_y)

                # out of bounds intentional
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore", UserWarning)
                    jacobian = self.ref2element_jacobian(index, quadrature_points)

                jacobians[index_x, index_y] = jacobian

        return jacobians

    def get_jacobian_dets(self) -> np.ndarray:
        """
        Compute the determinants of the Jacobians evaluated on
        all elements and all quadrature points.

        Returns:
            np.ndarray: Jacobian determinants.
                Of shape (num_elements_x+1, num_elements_y+1, num_quadrature_points).
        """

        jacobians = self.get_jacobian_tensors()
        jacobian_dets = np.linalg.det(jacobians)

        return jacobian_dets

    def get_jacobian_invdets(self) -> np.ndarray:  # noqa
        """
        Compute the inverse of the determinants of the Jacobians
        evaluated on all elements and all quadrature points.

        Returns:
            np.ndarray: Inverse Jacobian determinants.
                Of shape (num_elements_x+1, num_elements_y+1, num_quadrature_points).
        """

        jacobian_dets = self.get_jacobian_dets()
        jacobian_invdets = 1.0 / jacobian_dets

        return jacobian_invdets

    def __tca_default(self, oracle: Callable[[np.ndarray], np.ndarray]) -> TensorTrain:
        """
        Perform tensor train cross approximation for a given oracle function.
        The tensor shape is assumed to be [4] * mesh_size_exponent.

        Args:
            oracle (Callable[[np.ndarray], np.ndarray]): The oracle function to approximate.

        Returns:
            TensorTrain: The tensor train cross approximation of the oracle.
        """

        kwargs = self.tt_cross_config.to_dict()
        num_indices = kwargs.pop("num_anova_init")
        order = kwargs.pop("anova_order")
        tensor_shape = [4] * self.mesh_size_exponent
        train_indices = gen_teneva_indices(num_indices, tensor_shape)

        tt_init = anova_init_tensor_train(oracle, train_indices, order)
        tt_cross_cores = tensor_train_cross_approximation(oracle, tt_init, **kwargs)

        return TensorTrain([torch.tensor(core) for core in tt_cross_cores])

    def _cross_func(self, qindex: np.ndarray, xi_eta: np.ndarray) -> np.ndarray:
        """
        Compute the Jacobian for a given element index given in binary
        and reference coordinates.

        Args:
            qindex (np.ndarray): The quaternary index of the element.
                Of shape (num_indices, index_length) or (index_length,).
            xi_eta (np.ndarray): The reference coordinates in the quadrilateral.
                Of shape (1, 2) or (2,).

        Returns:
            np.ndarray: The Jacobian. Of shape (num_indices, 2, 2).

        Raises:
            ValueError: If the index map is not defined, if the reference coordinates are invalid
                or if the number of evaluation points is more than 1.
        """

        if self.index_map is None:
            raise ValueError("Index map is not defined.")

        xi_eta_ = ensure_2d(xi_eta)
        qindex_ = ensure_2d(qindex)
        self._validate_ref_coords(xi_eta_)

        if xi_eta_.shape[0] > 1:
            raise ValueError("Only one evaluation point is supported for TCA.")

        jacobians = []
        for idx in range(qindex_.shape[0]):
            single_bindex = np.array(qindex_[idx, :])
            index = self.index_map(single_bindex)
            jacobian = self.ref2element_jacobian(index, xi_eta_)[0]
            jacobians.append(jacobian)

        jac = np.stack(jacobians)
        return jac

    def plot_element(self, index: Tuple[int, int], num_points: int = 100) -> None:  # noqa
        """
        Plot the 2D points generated by the ref2element_map for a given index.

        Args:
            index (Tuple[int, int]): The 2D index of the element.
            num_points (int): The resolution of the grid for evaluation.
        """

        xi = np.linspace(-1, 1, num_points)
        eta = np.linspace(-1, 1, num_points)
        XI, ETA = np.meshgrid(xi, eta)

        # Flatten the grid for batch evaluation
        ref_coords = np.vstack([XI.ravel(), ETA.ravel()]).T

        # Evaluate the transformation map
        transformed_points = self.ref2element_map(index, ref_coords)

        # Extract X and Y coordinates
        X = transformed_points[:, 0]
        Y = transformed_points[:, 1]

        # Plot the 2D scatter plot of the mapped points
        plt.scatter(X, Y, c="red", marker="o", alpha=0.6)

        plt.title(f"2D Mapped Points for Index {index}")
        plt.xlabel("X")
        plt.ylabel("Y")
        plt.show()

    def plot(self, num_points: int = 100) -> None:
        """
        Plot the boundaries of all elements in the mesh with interpolated curves.

        Args:
            num_points (int): Number of points to sample along each curve.

        Warning:
            This method is not efficient for large meshes,
            call it with a small number of points for visualization purposes only.
        """
        xi_eta_edges = [
            np.column_stack((np.linspace(-1, 1, num_points), -1 * np.ones(num_points))),
            np.column_stack((np.ones(num_points), np.linspace(-1, 1, num_points))),
            np.column_stack((np.linspace(1, -1, num_points), np.ones(num_points))),
            np.column_stack((-1 * np.ones(num_points), np.linspace(1, -1, num_points))),
        ]

        for index_x in range(self.num_elements1d):
            for index_y in range(self.num_elements1d):
                for edge in xi_eta_edges:
                    physical_edge = np.array(
                        [
                            self.ref2element_map((index_x, index_y), xi_eta[np.newaxis, :])[0]
                            for xi_eta in edge
                        ]
                    )
                    plt.plot(
                        physical_edge[:, 0],
                        physical_edge[:, 1],
                        "b-",
                        label="Element Boundary" if (index_x, index_y) == (0, 0) else "",
                    )

        plt.axis("equal")
        plt.title("Mesh Plot")
        plt.xlabel("x")
        plt.ylabel("y")
        plt.show()

    def __repr__(self):
        return (
            f"SubdomainMesh2D(subdomain={self.subdomain},"
            f" quadrature_rule={self.quadrature_rule},"
            f" mesh_size_exponent={self.mesh_size_exponent},"
            f" num_points={self.num_points},"
            f" num_elements={self.num_elements})"
        )

    def _validate_idxs(self, index_x: int, index_y: int):
        """
        Validate the element indices.

        Args:
            index_x (int): The x-index of the element.
            index_y (int): The y-index of the element.

        Raises:
            ValueError: If the indices are out of bounds.

        Note:
            The indices are assumed to be zero-based.
            Indices are valid if 0 <= index_x, index_y <= num_elements1d.
            We include the upper bound to allow for the last "padded" element
            such that the total number of elements is a power of 2: (num_elements1d)**2.
        """
        if index_x < 0 or index_x >= self.num_elements1d:
            warnings.warn(
                f"Index x={index_x} is out of bounds [0, {self.num_elements1d})."
                f" The last element is a padded element."
            )
        if index_y < 0 or index_y >= self.num_elements1d:
            warnings.warn(
                f"Index y={index_y} is out of bounds [0, {self.num_elements1d})."
                f" The last element is a padded element."
            )

    def _validate_ref_coords(self, xi_eta: np.ndarray, tol: float = 1e-6):
        """
        Validate the reference coordinates.

        Args:
            xi_eta (np.ndarray): The reference coordinates in the quadrilateral.
                Of shape (num_points, 2).
            tol (float): The tolerance for the range check.

        Raises:
            ValueError: If the reference coordinates are out of bounds.
        """
        if not xi_eta.shape[1] == 2:
            raise ValueError("Reference coordinates must have shape (num_points, 2).")

        if not np.all(-1 - tol <= xi_eta) or not np.all(xi_eta <= 1 + tol):
            warnings.warn(
                f"Reference coordinates are not in the range [-1, 1]"
                f" within tolerance {tol}."
                " This behavior may be intentional when using tensorized Jacobians."
            )


class QuadMesh(SubdomainMesh2D):
    """
    Mesh for a quadrilateral subdomain.
    The Jacobians of quadrilateral meshes depend linearly on the element index.
    Hence, instead of using the tensor train cross approximation for a generic subdomain,
    we can represent the Jacobians for all elements and all quadrature points numerically
    exactly with a low-rank Tensor Train.

    Example:
    >>> from ttfemesh.domain import RectangleFactory
    >>> from ttfemesh.quadrature import GaussLegendre2D
    >>> from ttfemesh.mesh import QuadMesh

    >>> lower_left = (0, 0)
    >>> upper_right = (2, 1)
    >>> rectangle = RectangleFactory.create(lower_left, upper_right)
    >>> quadrature_rule = GaussLegendre2D()
    >>> mesh_size_exponent = 3
    >>> mesh = QuadMesh(rectangle, quadrature_rule, mesh_size_exponent)
    >>> mesh.plot()
    """

    def __init__(
        self,
        quad: Quad,
        quadrature_rule: QuadratureRule2D,
        mesh_size_exponent: int,
        tt_cross_config: Optional[TTCrossConfig] = None,
    ):
        """
        Initialize a quadrilateral mesh.

        Args:
            quad (Quad): The quadrilateral subdomain to mesh.
            quadrature_rule (QuadratureRule2D): The quadrature rule to use.
            mesh_size_exponent (int): The exponent of the discretization size.
                The discretization size is 2**(mesh_size_exponent) per dimension.
            tt_cross_config (Optional[TTCrossConfig]):
                Configuration for the tensor train cross approximation.
                If None, default configuration is used.
        """
        super().__init__(quad, quadrature_rule, mesh_size_exponent, tt_cross_config)
        self._tca_strategy = self.__linear_interpolation

    def __linear_interpolation(self, oracle: Callable[[np.ndarray], np.ndarray]) -> TensorTrain:
        """
        Perform linear interpolation for a given oracle function.

        Args:
            oracle (Callable[[np.ndarray], np.ndarray]): The oracle function to approximate.

        Returns:
            List[np.ndarray]: The tensor train cores of the linear interpolation.
        """

        def func(idx):
            return oracle(idx)[0]

        tt_interpolant = interpolate_linear2d(func, self.mesh_size_exponent)

        return tt_interpolant

    def __repr__(self):
        return (
            f"QuadMesh(subdomain={self.subdomain},"
            f" quadrature_rule={self.quadrature_rule},"
            f" mesh_size_exponent={self.mesh_size_exponent},"
            f" num_points={self.num_points},"
            f" num_elements={self.num_elements})"
        )



from abc import ABC, abstractmethod
from typing import Sequence

import matplotlib.pyplot as plt
import numpy as np

from ttfemesh.domain.curve import Curve, Line2D


class Subdomain(ABC):
    """
    Abstract base class for a subdomain in the domain.
    The subdomains are intended to "glue" more complex domains together.
    """

    @abstractmethod
    def _validate(self):  # pragma: no cover
        """Ensure that the subdomain is valid."""
        pass

    @abstractmethod
    def plot(self):  # pragma: no cover
        """Plot the subdomain and its boundaries."""
        pass


class Subdomain2D(Subdomain):
    """
    A 2D subdomain defined by 4 boundary curves.
    The curves must connect properly to form a closed subdomain.
    The start and end points of the curves must be ordered counter-clockwise.
    Note that the curves are considered to be ordered as follows:
    bottom (curve 0), right (curve 1), top (curve 2), left (curve 3).
    This is important for the boundary condition to work correctly.
    It may lead to confusion if, e.g., your curve 0 is visually
    the right edge of the domain.

    Example:
    >>> import numpy as np
    >>> from ttfemesh.domain import CircularArc2D, Line2D
    >>> from ttfemesh.domain import Subdomain2D

    >>> arc0 = CircularArc2D((0, 0), 1, np.pi/2., 0.5*np.pi)
    >>> line1 = Line2D((-1, 0), (1, -1))
    >>> line2 = Line2D((1, -1), (2, 1))
    >>> line3 = Line2D((2, 1), (0, 1))

    >>> subdomain = Subdomain2D([arc0, line1, line2, line3])
    >>> subdomain.plot()
    """

    def __init__(self, curves: Sequence[Curve]):
        """
        Initialize a 2D subdomain defined by 4 boundary curves.

        Args:
            curves (Sequence[Curve]): List of 4 boundary curves.
        """
        if len(curves) != 4:
            raise ValueError("A 2D subdomain must be defined by exactly 4 curves.")
        self.curves = curves
        self._validate()

    def get_curve(self, index: int) -> Curve:  # noqa
        """
        Get the curve at the specified index.

        Args:
            index (int): Index of the curve.

        Returns:
            Curve: The curve at the specified index.
        """

        if index < 0 or index >= 4:
            raise ValueError("Curve index must be in the range [0, 3].")

        return self.curves[index]

    def _check_connect(self, tol: float = 1e-6):
        """
        Ensure that the curves connect properly.

        Args:
            tol (float): Tolerance for point-wise comparison. Default is 1e-6.
        """

        for i in range(4):
            end = self.curves[i].get_end()
            next_start = self.curves[(i + 1) % 4].get_start()
            if not np.allclose(end, next_start, atol=tol):
                raise ValueError(f"Curves {i} and {(i + 1) % 4} do not connect properly.")

    def _check_orientation(self):
        """Check if the curves defining the domain are ordered counter-clockwise."""

        points = [curve.get_start() for curve in self.curves]
        signed_area = 0.0
        for i in range(len(points) - 1):
            x1, y1 = points[i]
            x2, y2 = points[i + 1]
            signed_area += x1 * y2 - x2 * y1

        if signed_area < 0:
            raise ValueError("The start points of curves are not ordered counter-clockwise.")

    def _validate(self):
        self._check_connect()
        self._check_orientation()

    def plot(self, num_points: int = 100):
        for i, curve in enumerate(self.curves):
            t_vals = np.linspace(-1, 1, num_points)
            points = curve.evaluate(t_vals)
            plt.plot(points[:, 0], points[:, 1], label=f"curve {i}")

        plt.title("Subdomain")
        plt.axis("equal")
        plt.show()

    def __repr__(self):
        points = [curve.get_start() for curve in self.curves]
        return f"Subdomain2D(points={points})"


class Quad(Subdomain2D):
    """
    Quadrilateral subdomain defined by 4 boundary lines.
    Note that the lines are considered to be ordered as follows:
    bottom (line 0), right (line 1), top (line 2), left (line 3).
    This is important for the boundary condition to work correctly.
    It may lead to confusion if, e.g., your line 0 is visually
    the right edge of the domain.

    Recommend using the QuadFactory or the RectangleFactory to create a quadrilateral subdomain.

    Example:
    >>> from ttfemesh.domain import Line2D
    >>> from ttfemesh.domain import Quad

    >>> line0 = Line2D((0, 1), (-1, 0))
    >>> line1 = Line2D((-1, 0), (1, -1))
    >>> line2 = Line2D((1, -1), (2, 1))
    >>> line3 = Line2D((2, 1), (0, 1))

    >>> subdomain = Quad([line0, line1, line2, line3])
    >>> subdomain.plot()
    """

    def __init__(self, curves: Sequence[Line2D]):
        super().__init__(curves)

    def __repr__(self):
        points = [curve.get_start() for curve in self.curves]
        return f"Quad(points={points})"



from .boundary_condition import DirichletBoundary2D
from .curve import CircularArc2D, Curve, Line2D, ParametricCurve2D
from .domain import Domain, Domain2D
from .subdomain import Quad, Subdomain, Subdomain2D
from .subdomain_connection import CurveConnection2D, VertexConnection2D
from .subdomain_factory import QuadFactory, RectangleFactory

__all__ = [
    "Curve",
    "Line2D",
    "CircularArc2D",
    "ParametricCurve2D",
    "Subdomain",
    "Subdomain2D",
    "Quad",
    "RectangleFactory",
    "QuadFactory",
    "VertexConnection2D",
    "CurveConnection2D",
    "DirichletBoundary2D",
    "Domain",
    "Domain2D",
]



from .interpolate import interpolate_linear2d
from .meshgrid import map2canonical2d, range_meshgrid2d, zmeshgrid2d
from .numeric import unit_vector_binary_tt
from .operations import levelwise_kron, transpose_kron, zorder_kron
from .tensor_cross import (
    TTCrossConfig,
    anova_init_tensor_train,
    error_on_indices,
    error_on_random_indices,
    gen_teneva_indices,
    tensor_train_cross_approximation,
)

__all__ = [
    "anova_init_tensor_train",
    "gen_teneva_indices",
    "TTCrossConfig",
    "tensor_train_cross_approximation",
    "error_on_indices",
    "error_on_random_indices",
    "zorder_kron",
    "transpose_kron",
    "levelwise_kron",
    "range_meshgrid2d",
    "zmeshgrid2d",
    "map2canonical2d",
    "interpolate_linear2d",
    "unit_vector_binary_tt",
]



from typing import Callable, List, Optional

import numpy as np
import teneva


class TTCrossConfig:
    """
    Configuration class for the tensor train cross approximation algorithm.

    Args:
        cache (dict, optional): Cache for storing requested function values.
        info (dict, optional): Stores TTCross run information.
        num_sweeps (int, optional): Number of sweeps for DMRG. Defaults to 10.
        rel_stagnation_tol (float, optional): Relative stagnation tolerance. Defaults to 1e-4.
        max_func_calls (Optional[int], optional): Maximum number of function calls.
            Defaults to None.
        cache_calls_factor (int, optional): If the number of calls to cache is this factor
            times larger than number of function calls, TTCross stops. Defaults to 20.
        num_anova_init (int, optional): Number of training indices for ANOVA initialization.
            Defaults to 1000.
        anova_order (int, optional): Order of the ANOVA decomposition. Defaults to 2.
        verbose (bool, optional): Verbose output. Defaults to False.
    """

    def __init__(
        self,
        cache: Optional[dict] = None,
        info: Optional[dict] = None,
        num_sweeps: int = 10,
        rel_stagnation_tol: float = 1e-4,
        max_func_calls: Optional[int] = None,
        cache_calls_factor: int = 5,
        num_anova_init: int = 1000,
        anova_order: int = 2,
        verbose: bool = False,
    ):

        self.cache = cache
        self.info = info
        self.num_sweeps = num_sweeps
        self.rel_stagnation_tol = rel_stagnation_tol
        self.max_func_calls = max_func_calls
        self.cache_calls_factor = cache_calls_factor
        self.num_anova_init = num_anova_init
        self.anova_order = anova_order
        self.verbose = verbose

    def to_dict(self):
        """
        Convert all attributes of the configuration to a dictionary that is passed to
        teneva

        Returns:
            dict: A dictionary representation of the configuration.
        """
        kwargs = {
            "cache": self.cache,
            "info": self.info,
            "nswp": self.num_sweeps,
            "e": self.rel_stagnation_tol,
            "log": self.verbose,
            "m": self.max_func_calls,
            "m_cache_scale": self.cache_calls_factor,
            "num_anova_init": self.num_anova_init,
            "anova_order": self.anova_order,
        }

        return kwargs


def gen_teneva_indices(num_indices: int, tensor_shape: List[int]) -> np.ndarray:
    """
    Generate random indices for a tensor of shape tensor_shape.

    Args:
        num_indices (int): Number of indices to generate.
        tensor_shape (List[int]): Shape of the tensor.

    Returns:
        np.ndarray: Random indices of shape (num_indices, len(tensor_shape)).
    """
    idxs = np.vstack([np.random.choice(k, num_indices) for k in tensor_shape]).T
    return idxs


def anova_init_tensor_train(
    oracle: Callable[[np.ndarray], np.ndarray], train_indices: np.ndarray, order: int = 2
) -> List[np.ndarray]:
    """
    Initialize the tensor train with the ANOVA decomposition of the training data.

    Args:
        oracle (Callable[[np.ndarray], np.ndarray]): Oracle function.
        train_indices (np.ndarray): Training indices.
        order (int, optional): Order of the ANOVA decomposition. Defaults to 2.

    Returns:
        List[np.ndarray]: List of TT-cores for the ANOVA decomposition of the training data.
    """
    ytrain = oracle(train_indices)
    yanova = teneva.anova(train_indices, ytrain, order=order)
    return yanova


def tensor_train_cross_approximation(
    oracle: Callable[[np.ndarray], np.ndarray], tt_init: List[np.ndarray], **kwargs
) -> List[np.ndarray]:
    """
    Approximate the tensor train with the cross approximation algorithm.

    Args:
        oracle (Callable[[np.ndarray], np.ndarray]): Oracle function.
        tt_init (List[np.ndarray]): Initial tensor train.
        **kwargs: Additional keyword arguments for the cross approximation algorithm.

    Returns:
        List[np.ndarray]: List of TT-cores for the approximated tensor train.
    """

    return teneva.cross(oracle, tt_init, **kwargs)


def error_on_indices(
    oracle: Callable[[np.ndarray], np.ndarray],
    approx_tt: List[np.ndarray],
    test_indices: np.ndarray,
) -> float:
    """
    Test the accuracy of the approximated tensor train.

    Args:
        oracle (Callable[[np.ndarray], np.ndarray]): Oracle function.
        approx_tt (List[np.ndarray]): Approximated tensor train cores.
        test_indices (np.ndarray): Test indices.

    Returns:
        float: Relative error of the approximated tensor train.
    """
    ytest = oracle(test_indices)
    error = teneva.accuracy_on_data(approx_tt, test_indices, ytest)
    return error


def error_on_random_indices(
    oracle: Callable[[np.ndarray], np.ndarray],
    approx_tt: List[np.ndarray],
    num_test_indices: int,
    tensor_shape: List[int],
) -> float:
    """
    Test the accuracy of the approximated tensor train with random test indices.

    Args:
        oracle (Callable[[np.ndarray], np.ndarray]): Oracle function.
        approx_tt (List[np.ndarray]): Approximated tensor train cores.
        num_test_indices (int): Number of test indices.
        tensor_shape (List[int]): Shape of the tensor.

    Returns:
        float: Relative error of the approximated tensor train.
    """
    test_indices = gen_teneva_indices(num_test_indices, tensor_shape)
    return error_on_indices(oracle, approx_tt, test_indices)



import numpy as np
import torchtt as tntt

from ttfemesh.tt_tools.operations import zorder_kron
from ttfemesh.types import TensorTrain


def zmeshgrid2d(X: TensorTrain, Y: TensorTrain) -> TensorTrain:
    """
    Compute the meshgrid of two TT-tensors using the Z-ordering.

    Args:
        X (TensorTrain): First TT-tensor.
        Y (TensorTrain): Second TT-tensor.

    Returns:
        TensorTrain resulting from the meshgrid of X and Y.
    """

    ones_x = tntt.ones(X.N)
    ones_y = tntt.ones(Y.N)
    XX = zorder_kron(X, ones_y)
    YY = zorder_kron(ones_x, Y)

    return XX, YY


def range_meshgrid2d(mesh_size_exponent: int) -> TensorTrain:
    """
    Compute the meshgrid corresponding to X and Y tensors counting from 0 to 2**d.

    Args:
        mesh_size_exponent (int): Exponent of 1D grid size.

    Returns:
        TensorTrain resulting from the meshgrid of two index tensors.
    """

    range = tntt._extras.xfun([2] * mesh_size_exponent)
    return zmeshgrid2d(range, range)


def map2canonical2d(mesh_size_exponent: int) -> np.ndarray:
    """
    Computes a vector where the i-th element is the index of the i-th element
    in the z-ordering of the meshgrid.
    The canonical ordering for a 2D grid indexed by (i, j) is computed as
    i + j * 2**d and does not correspond to the z-ordering of (i, j).
    When a TT-tensor in the z-ordering is reshaped to a vector,
    the order of the elements is given by the vector returned by this function.
    This effectively allows to map the z-ordering back to the canonical ordering.
    Useful for debugging and testing.

    Args:
        mesh_size_exponent (int): Exponent of 1D grid size.

    Returns:
        np.ndarray: Vector of indices mapping z-ordering to canonical ordering.

    Example:
        >>> array = tt.full().flatten("F")
        >>> zmap = map2canonical2d(3)
        >>> canonical_array = np.empty_like(array)
        >>> canonical_array[zmap] = array
    """

    meshgrid = range_meshgrid2d(mesh_size_exponent)
    XX, YY = meshgrid
    vecx = np.array(XX.full()).flatten("F")
    vecy = np.array(YY.full()).flatten("F")
    z2tuples = list(zip(vecx, vecy))
    xlen = 2**mesh_size_exponent

    def index_map(pair):
        i, j = pair
        idx = int(j * xlen + i)
        return idx

    zmap = [index_map(pair) for pair in z2tuples]

    return np.array(zmap)



from abc import ABC, abstractmethod
from collections import defaultdict
from typing import Dict, List, Tuple

from ttfemesh.domain.subdomain import Subdomain2D


class BoundaryCondition(ABC):
    """
    Abstract base class for boundary conditions.
    """

    @abstractmethod
    def validate(self, *args, **kwargs):  # pragma: no cover
        """
        Validate the boundary condition with respect to the subdomains.
        """
        pass

    @abstractmethod
    def group_by_subdomain(self) -> Dict[int, List[int]]:  # pragma: no cover
        """
        Group the boundary conditions by subdomain.

        Returns:
            Dict[int, List[int]]: A dictionary where the keys are subdomain indices,
                                  and the values are lists of curve or face indices
                                  with the boundary condition.
        """
        pass


class DirichletBoundary2D(BoundaryCondition):
    """
    Implements a Dirichlet boundary condition for a 2D curve.
    Boundary values are implicitly assumed to be zero.

    Example:
    >>> from ttfemesh.domain import RectangleFactory, CurveConnection2D, VertexConnection2D
    >>> from ttfemesh.domain import DirichletBoundary2D, Domain2D
    >>> lower_left = (0, 0)
    >>> upper_right = (2, 1)
    >>> rectangle1 = RectangleFactory.create(lower_left, upper_right)

    >>> lower_left = (2, 0)
    >>> upper_right = (3, 1)
    >>> rectangle2 = RectangleFactory.create(lower_left, upper_right)

    >>> lower_left = (-2, 1)
    >>> upper_right = (0, 2)
    >>> rectangle3 = RectangleFactory.create(lower_left, upper_right)

    >>> domain_idxs = [0, 1]
    >>> curve_idxs = [1, 3]
    >>> edge = CurveConnection2D(domain_idxs, curve_idxs)

    >>> vertex_idxs = [(0, 3, "start"), (2, 0, "end")]
    >>> vertex = VertexConnection2D(vertex_idxs)

    >>> bc = DirichletBoundary2D([(1, 1), (2, 3)])

    >>> domain = Domain2D([rectangle1, rectangle2, rectangle3], [edge, vertex], bc)
    >>> domain.plot()
    """

    def __init__(self, boundary: List[Tuple[int, int]]):
        """
        Initialize the Dirichlet boundary condition.

        Args:
            boundary (List[Tuple[int, int]]):
                List of subdomain and curve indices for the Dirichlet boundary.
        """
        self.boundary = boundary

    def validate(self, subdomains: List[Subdomain2D]):
        """
        Validate the boundary condition. Ensure that the specified curve exists.

        Args:
            subdomains (List[Subdomain]): List of subdomains in the domain.
        """
        for subdomain_idx, curve_idx in self.boundary:
            if subdomain_idx >= len(subdomains):
                raise ValueError(f"Subdomain index {subdomain_idx} out of range.")
            if curve_idx >= len(subdomains[subdomain_idx].curves):
                raise ValueError(
                    f"Curve index {curve_idx} out of range" f" for subdomain {subdomain_idx}."
                )

    def num_bcs(self) -> int:
        """
        Get the number of curves with imposed boundary conditions.

        Returns:
            int: Number of curves with boundary conditions.
        """
        return len(self.boundary)

    def group_by_subdomain(self) -> Dict[int, List[int]]:
        """
        Group the boundary conditions by subdomain.

        Returns:
            Dict[int, List[int]]: A dictionary where the keys are subdomain indices,
                                  and the values are lists of curve indices with
                                  the boundary condition.
        """
        grouped = defaultdict(list)
        for subdomain, curve in self.boundary:
            grouped[subdomain].append(curve)

        return dict(grouped)

    def __repr__(self):
        return f"DirichletBoundary2D(num_bcs={self.num_bcs()})"



from abc import ABC, abstractmethod
from typing import Optional, Tuple

import numpy as np
import numpy.polynomial.legendre as leg


class QuadratureRule(ABC):
    """Abstract base class for a quadrature rule."""

    @abstractmethod
    def get_points_weights(self) -> Tuple[np.ndarray, np.ndarray]:  # pragma: no cover
        """Retrieve the quadrature points and weights."""
        pass

    @staticmethod
    @abstractmethod
    def compute_points_weights(self) -> Tuple[np.ndarray, np.ndarray]:  # pragma: no cover
        """Compute the quadrature points and weights."""
        pass

    @property
    @abstractmethod
    def dimension(self) -> int:  # pragma: no cover
        """The number of dimensions of the quadrature rule."""
        pass


class QuadratureRule2D(QuadratureRule):
    """Abstract base class for a 2D quadrature rule."""

    @property
    def dimension(self) -> int:
        """The number of dimensions of the quadrature rule."""
        return 2


class GaussLegendre(QuadratureRule):
    """
    Implements Gauss-Legendre quadrature on [-1, 1]^(dimension).

    Example:
    >>> from ttfemesh.quadrature import GaussLegendre

    >>> order = 3
    >>> dim = 3
    >>> qrule = GaussLegendre(order, dim)

    >>> points, weights = qrule.get_points_weights()

    >>> print(points)
    >>> print(weights)
    """

    def __init__(self, order: int = 1, dimension: int = 1):
        """
        Initialize the Gauss-Legendre quadrature rule.

        Args:
            order (int): The order of the quadrature rule. Must be greater than 0 (default is 1).
            dimension (int): The number of dimensions (default is 1).
        """
        self.order = order
        self._dimension = dimension
        self.points: Optional[np.ndarray] = None
        self.weights: Optional[np.ndarray] = None

    @property
    def dimension(self) -> int:
        """The number of dimensions of the quadrature rule."""
        return self._dimension

    def get_points_weights(self) -> Tuple[np.ndarray, np.ndarray]:
        """
        Get the quadrature points and weights.

        Returns:
            Tuple[np.ndarray, np.ndarray]: Quadrature points and weights.
        """
        if self.points is None or self.weights is None:
            self.points, self.weights = self.compute_points_weights(self.order, self.dimension)
        return self.points, self.weights

    @staticmethod
    def compute_points_weights(order: int, dimension: int = 1) -> Tuple[np.ndarray, np.ndarray]:
        """
        Compute the Gauss-Legendre quadrature points and weights.

        Args:
            order (int): The order of the quadrature rule. Must be greater than 0.
            dimension (int): The number of dimensions (default is 1).

        Returns:
            Tuple[np.ndarray, np.ndarray]: Quadrature points and weights.
        """
        points_1d, weights_1d = leg.leggauss(order)

        if dimension == 1:
            return points_1d[:, None], weights_1d

        # Tensor product for multi-dimensional quadrature
        grid = np.meshgrid(*[points_1d] * dimension, indexing="ij")
        points = np.stack([g.flatten() for g in grid], axis=-1)

        weights = np.prod(np.meshgrid(*[weights_1d] * dimension, indexing="ij"), axis=0)
        return points, weights.flatten()

    def __repr__(self) -> str:
        """String representation of the Gauss-Legendre quadrature rule."""
        return f"Gauss-Legendre Quadrature Rule (order={self.order}, dimension={self.dimension})"


class GaussLegendre2D(GaussLegendre):
    """
    Implements Gauss-Legendre quadrature on [-1, 1]^2.

    Example:
    >>> from ttfemesh.quadrature import GaussLegendre2D

    >>> order = 3
    >>> qrule2D = GaussLegendre2D(order)

    >>> points2D, weights2D = qrule2D.get_points_weights()

    >>> print(points2D)
    >>> print(weights2D)
    """

    def __init__(self, order: int = 1):
        """
        Initialize the 2D Gauss-Legendre quadrature rule.

        Args:
            order (int): The order of the quadrature rule. Must be greater than 0 (default is 1).
        """
        super().__init__(order, dimension=2)

    def __repr__(self) -> str:
        """String representation of the 2D Gauss-Legendre quadrature rule."""
        return f"2D Gauss-Legendre Quadrature Rule (order={self.order})"



from enum import Enum, auto
from typing import TypeAlias

import torchtt

TensorTrain: TypeAlias = torchtt.TT


class BoundarySide2D(Enum):
    """
    Enum for the sides of a 2D domain.
    Note that the sides are considered to be ordered as follows:
    bottom (side 0), right (side 1), top (side 2), left (side 3).
    This is important for the boundary condition to work correctly.
    It may lead to confusion if, e.g., your side 0 is visually
    the right edge of the domain.
    """

    BOTTOM = 0
    RIGHT = auto()
    TOP = auto()
    LEFT = auto()


class BoundaryVertex2D(Enum):
    """
    Enum for the vertices of a 2D domain.
    Note that the vertices are considered to be ordered as follows:
    bottom_left (vertex 0), bottom_right (vertex 1), top_right (vertex 2), top_left (vertex 3).
    This is important for the boundary condition to work correctly.
    It may lead to confusion if, e.g., your vertex 0 is visually
    the bottom left corner of the domain.
    """

    BOTTOM_LEFT = 0
    BOTTOM_RIGHT = auto()
    TOP_RIGHT = auto()
    TOP_LEFT = auto()



"""
TTFEMesh is a Python package for finite element methods using tensor train
decompositions.

The package implements a framework for creating and manipulating finite element
meshes in the tensor train format. This provides the backbone for solving partial
differential equations using TT decompositions. The package is designed
to be flexible and extensible, allowing users to create their own finite element
meshes and build their own solvers.

The package is built on top of the `torchtt` package, which provides a framework
for handling TT decompositions, but this might change in the future.
This package is currently in a prototype stage, and the API is subject to change.

See the documentation for more details.

Example usage:
    >>> from ttfemesh.domain import RectangleFactory
    >>> from ttfemesh.quadrature import GaussLegendre2D
    >>> from ttfemesh.mesh import SubdomainMesh2D

    >>> lower_left = (0, 0)
    >>> upper_right = (2, 1)
    >>> rectangle = RectangleFactory.create(lower_left, upper_right)
    >>> quadrature_rule = GaussLegendre2D(order=1)
    >>> mesh_size_exponent = 3
    >>> mesh = SubdomainMesh2D(rectangle, quadrature_rule, mesh_size_exponent)

    >>> mesh.plot()
"""

from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version("ttfemesh")
except PackageNotFoundError:
    __version__ = "unknown"



from setuptools import setup, find_packages

ascii_logo = """
████████╗████████╗███████╗███████╗███╗   ███╗
╚══██╔══╝╚══██╔══╝██╔════╝██╔════╝████╗ ████║ 
   ██║      ██║   █████╗  █████╗  ██╔████╔██║ 
   ██║      ██║   ██╔══╝  ██╔══╝  ██║╚██╔╝██║ 
   ██║      ██║   ██║     ███████╗██║ ╚═╝ ██║ 
   ╚═╝      ╚═╝   ╚═╝     ╚══════╝╚═╝     ╚═╝ 
"""

print(ascii_logo)

requirements = [
    requirement.strip() for requirement in open('requirements.txt').readlines()
]

setup(
    name="ttfemesh",
    version="0.1.0",
    description="A Python library for tensor train-based finite element meshing.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="Mazen Ali",
    author_email="mazen.ali90@gmail.com",
    url="https://github.com/MazenAli/tt-femesh",
    packages=find_packages(),
    install_requires=requirements,
    python_requires=">=3.9, <3.13",
    license="MIT",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Topic :: Software Development :: Libraries",
    ],
    project_urls={
        "Source": "https://github.com/MazenAli/tt-femesh",
        "Documentation": "https://github.com/MazenAli/tt-femesh",
    },
    keywords="tensor train finite elements meshing simulation",
)



import warnings
from abc import ABC, abstractmethod
from typing import Callable, Union

import numpy as np

from ttfemesh.utils.array import ensure_1d


class Curve(ABC):
    """
    Abstract base class for a curve in the domain.
    Defines the interface for all curve implementations.
    """

    def get_start(self) -> np.ndarray:
        """Get the start point of the curve."""
        return self.evaluate(-1.0)[0]

    def get_end(self) -> np.ndarray:
        """Get the end point of the curve."""
        return self.evaluate(1.0)[0]

    def __call__(self, *args, **kwargs) -> np.ndarray:
        return self.evaluate(*args, **kwargs)

    def _validate(self, t: Union[np.ndarray, float], tol: float = 1e-6):
        """
        Ensure that parameter values are in the interval [-1, 1] within a specified tolerance.

        Args:
            t (np.ndarray): Array of parameter values to validate.
            tol (float): Tolerance for boundary checks.

        Raises:
            ValueError: If any parameter values are outside the interval [-1, 1] (with tolerance).
        """

        t_ = ensure_1d(t)

        if not np.all(-1 - tol <= t_) or not np.all(t_ <= 1 + tol):
            warnings.warn(
                f"Parameter values are not in the interval [-1, 1]"
                f" within tolerance {tol}."
                f" May lead to unexpected behavior."
            )

    @abstractmethod
    def evaluate(self, t: Union[np.ndarray, float]) -> np.ndarray:  # pragma: no cover
        """
        Evaluate the curve at parameter values t.

        Args:
            t (Union[np.ndarray, float]): Array of or scalar parameter values in [-1, 1].

        Returns:
            np.ndarray: Array of shape (len(t), 2) with (x, y) coordinates.
        """
        pass

    @abstractmethod
    def tangent(self, t: Union[np.ndarray, float]) -> np.ndarray:  # pragma: no cover
        """
        Compute the tangent vector (not normalized) to the curve at parameter values t.

        Args:
            t (Union[np.ndarray, float]): Array of or scalar parameter values in [-1, 1].

        Returns:
            np.ndarray: Array of shape (len(t), 2) with tangent vectors.
        """
        pass

    def equals(self, other: "Curve", num_points: int = 100, tol: float = 1e-6) -> bool:
        """
        Check if two curves are approximately equal by sampling.

        Args:
            other (Curve): Another curve to compare.
            num_points (int): Number of points to sample along the curve.
            tol (float): Tolerance for point-wise comparison.

        Returns:
            bool: True if the curves are approximately equal, False otherwise.
        """
        reverse = False
        start = self.get_start()
        start_other = other.get_start()

        if not np.allclose(start, start_other, atol=tol):
            start_other = other.get_end()
            if not np.allclose(start, start_other, atol=tol):
                return False
            reverse = True

        ts = np.linspace(-1, 1, num_points)
        for t in ts:
            t_other = t if not reverse else -t
            if not np.allclose(self.evaluate(t), other.evaluate(t_other), atol=tol):
                return False
        return True


class Line2D(Curve):
    """
    A line segment in 2D space.

    Example:
        >>> from ttfemesh.domain import Line2D
        >>> from ttfemesh.utils import plot_curve_with_tangents
        >>> line = Line2D((0, 0), (1, 1))
        >>> print(line)
        >>> plot_curve_with_tangents(line, "Line")
    """

    def __init__(self, start: tuple[float, float], end: tuple[float, float]):
        """
        Initialize a line segment from `start` to `end`.

        Args:
            start (tuple): Coordinates of the start point (x, y).
            end (tuple): Coordinates of the end point (x, y).
        """
        self.start = np.array(start)
        self.end = np.array(end)

    def evaluate(self, t: Union[np.ndarray, float]) -> np.ndarray:
        t_ = ensure_1d(t)
        self._validate(t_)

        return np.outer((1 - t_) * 0.5, self.start) + np.outer((1 + t_) * 0.5, self.end)

    def tangent(self, t: Union[np.ndarray, float]) -> np.ndarray:
        t_ = ensure_1d(t)
        self._validate(t_)

        return np.tile(0.5 * (self.end - self.start), (len(t_), 1))

    def __repr__(self):
        return f"Line2D(start={tuple(self.start)}, end={tuple(self.end)})"


class CircularArc2D(Curve):
    """
    A circular arc in 2D space.

    Example:
        >>> import numpy as np
        >>> from ttfemesh.domain import CircularArc2D
        >>> from ttfemesh.utils import plot_curve_with_tangents
        >>> circular_arc = CircularArc2D((0, 0), 1, np.pi/2., 0.5*np.pi)
        >>> print(circular_arc)
        >>> plot_curve_with_tangents(circular_arc, "Circular Arc")
    """

    def __init__(
        self,
        center: tuple[float, float],
        radius: float,
        start_angle: float = 0.0,
        angle_sweep: float = np.pi,
    ):
        """
        Initialize a circular arc defined by a center, radius, and angle sweep.

        Args:
            center (tuple): Coordinates of the center (x, y).
            radius (float): Radius of the half-circle.
            start_angle (float): Starting angle in radians. Default is 0.
            angle_sweep (float): Angle sweep in radians. Default is π.
        """
        self.center = np.array(center)
        self.radius = radius
        self.start_angle = start_angle
        self.angle_sweep = angle_sweep

    def evaluate(self, t: Union[np.ndarray, float]) -> np.ndarray:
        t_ = ensure_1d(t)
        self._validate(t_)

        angle = self.start_angle + 0.5 * (t_ + 1) * self.angle_sweep
        x = self.center[0] + self.radius * np.cos(angle)
        y = self.center[1] + self.radius * np.sin(angle)
        return np.stack((x, y), axis=-1)

    def tangent(self, t: Union[np.ndarray, float]) -> np.ndarray:
        t_ = ensure_1d(t)
        self._validate(t_)

        angle = self.start_angle + 0.5 * (t_ + 1) * self.angle_sweep
        mul_factor = 0.5 * self.angle_sweep * self.radius
        dx = -np.sin(angle) * mul_factor
        dy = np.cos(angle) * mul_factor
        tangent = np.stack((dx, dy), axis=-1)
        return tangent

    def __repr__(self):
        return (
            f"CircularArc2D(center={tuple(self.center)}, "
            f"radius={self.radius}, start_angle={self.start_angle}, "
            f"angle_sweep={self.angle_sweep})"
        )


class ParametricCurve2D(Curve):
    """
    A parametric curve in 2D space.

    Example:
        >>> from ttfemesh.domain import ParametricCurve2D
        >>> from ttfemesh.utils import plot_curve_with_tangents
        >>> parametric_curve = ParametricCurve2D(
        ...     lambda t: np.sin(t * np.pi),
        ...     lambda t: np.cos(t * np.pi)
        ... )
        >>> print(parametric_curve)
        >>> plot_curve_with_tangents(parametric_curve, "Parametric Curve")
    """

    def __init__(
        self, x_func: Callable[[np.ndarray], np.ndarray], y_func: Callable[[np.ndarray], np.ndarray]
    ):
        """
        Initialize a parametric curve defined by functions x(t) and y(t).
        Uses a finite difference approximation to compute the tangent.

        Args:
            x_func (Callable[[np.ndarray], np.ndarray]): Function x(t) where t is in [-1, 1].
            y_func (Callable[[np.ndarray], np.ndarray]): Function y(t) where t is in [-1, 1].
        """
        self.x_func = x_func
        self.y_func = y_func

    def evaluate(self, t: Union[np.ndarray, float]) -> np.ndarray:
        t_ = ensure_1d(t)
        self._validate(t_)

        x = self.x_func(t_)
        y = self.y_func(t_)
        return np.stack((x, y), axis=-1)

    def tangent(self, t: Union[np.ndarray, float]) -> np.ndarray:
        """
        Compute the tangent vector using a finite difference approximation.

        Args:
            t (Union[np.ndarray, float]): Array of or scalar parameter values in [-1, 1].

        Returns:
            np.ndarray: Array of shape (len(t), 2) with tangent.
        """
        t_ = ensure_1d(t)
        self._validate(t_)

        dt = 1e-5
        dx = (self.x_func(t_ + dt) - self.x_func(t_)) / dt
        dy = (self.y_func(t_ + dt) - self.y_func(t_)) / dt
        tangent = np.stack((dx, dy), axis=-1)
        return tangent

    def __repr__(self):
        return f"ParametricCurve2D(x_func={self.x_func}, y_func={self.y_func})"



from typing import Tuple

import numpy as np
import torch

from ttfemesh.types import BoundarySide2D, BoundaryVertex2D, TensorTrain


def bindex2dtuple(bindex: np.ndarray) -> Tuple[int, int]:
    """
    Convert a binary index to a 2D tuple.
    Implements the mapping (i0, j0, i1, j1, ...) -> (i, j), where i = i0 + 2*i1 + 4*i2 + ...
    and j = j0 + 2*j1 + 4*j2 + ...
    This is the common little-endian convention in QTT literature.

    Args:
        bindex (np.ndarray): A binary index of shape (2 * num_bits1d,).

    Returns:
        Tuple[int, int]: A 2D tuple.

    Raises:
        ValueError: If bindex is not a 1D array, does not contain an even number of elements, or
            contains values other than 0 or 1.
    """

    shape = bindex.shape
    if len(shape) != 1:
        raise ValueError(f"Invalid shape ({shape}) for binary index. Expected 1D array.")

    if shape[0] % 2 != 0:
        raise ValueError("Binary index must have even number of elements.")

    if not np.all((bindex == 0) | (bindex == 1)):
        raise ValueError("Binary index must contain only 0s and 1s.")

    i_bits = bindex[::2]
    j_bits = bindex[1::2]

    i = np.sum(i_bits * (2 ** np.arange(len(i_bits))))
    j = np.sum(j_bits * (2 ** np.arange(len(j_bits))))

    return (i, j)


def qindex2dtuple(index: np.ndarray) -> Tuple[int, int]:
    """
    Convert a quaternary index to a 2D tuple.
    The ordering of the quaternary index is assumed to be
    (0, 1, 2, 3) -> ((0, 0), (1, 0), (0, 1), (1, 1)), i.e., column-major with
    index = i + 2*j.
    This is consistent with the little-endian ordering of the binary index
    commonly used in QTT literature.

    Args:
        index (np.ndarray): An index of shape (num_quats,) with values in {0, 1, 2, 3}.

    Returns:
        Tuple[int, int]: A 2D tuple.

    Raises:
        ValueError: If index is not a 1D array or contains values other than {0, 1, 2, 3}.
    """

    shape = index.shape
    if len(shape) != 1:
        raise ValueError(f"Invalid shape ({shape}) for index. Expected 1D array.")

    if not np.all((0 <= index) & (index <= 3)):
        raise ValueError("Index must contain only values in {0, 1, 2, 3}.")

    i = index % 2
    j = index // 2

    binary_index = np.empty(2 * len(index), dtype=int)
    binary_index[::2] = i
    binary_index[1::2] = j

    return bindex2dtuple(binary_index)


def side_concatenation_core(side: BoundarySide2D) -> np.ndarray:
    """
    Get the TT-core for concatenation of a boundary side.
    See Section 5.2 of arXiv:1802.02839 for details.

    Args:
        side (BoundarySide2D): The boundary side.

    Returns:
        np.ndarray: The TT-core for concatenation of the boundary side.
    """

    side_values = {
        BoundarySide2D.BOTTOM: 0,
        BoundarySide2D.RIGHT: 0,
        BoundarySide2D.TOP: 0,
        BoundarySide2D.LEFT: 0,
    }
    side_values[side] = 1
    B, R, T, L = (
        side_values[BoundarySide2D.BOTTOM],
        side_values[BoundarySide2D.RIGHT],
        side_values[BoundarySide2D.TOP],
        side_values[BoundarySide2D.LEFT],
    )

    core = np.array([[B, R, L, T], [L, B, T, R]]).reshape([1, 2, 4, 1])

    return core


def vertex_concatenation_core(vertex: BoundaryVertex2D) -> np.ndarray:
    """
    Get the TT-core for concatenation of a vertex.
    See Section 5.2 of arXiv:1802.02839 for details.

    Returns:
        np.ndarray: The TT-core for concatenation of a vertex.
    """

    vertex_values = {
        BoundaryVertex2D.BOTTOM_LEFT: 0,
        BoundaryVertex2D.BOTTOM_RIGHT: 0,
        BoundaryVertex2D.TOP_RIGHT: 0,
        BoundaryVertex2D.TOP_LEFT: 0,
    }
    vertex_values[vertex] = 1
    BL, BR, TR, TL = (
        vertex_values[BoundaryVertex2D.BOTTOM_LEFT],
        vertex_values[BoundaryVertex2D.BOTTOM_RIGHT],
        vertex_values[BoundaryVertex2D.TOP_RIGHT],
        vertex_values[BoundaryVertex2D.TOP_LEFT],
    )

    core = np.array([BL, BR, TL, TR]).reshape([1, 1, 4, 1])

    return core


def concat_core2tt(core: np.ndarray, length: int, exchanged: bool = False) -> TensorTrain:
    """
    Convert a TT-core to a TensorTrain.
    The rank-1 TT-core is assumed to be in the format (1, m, n, 1).

    Args:
        core (np.ndarray): The TT-core.
        length (int): The length of the TT.
        exchanged (bool): Whether the rows of the TT-core are exchanged.

    Returns:
        TensorTrain: The TT-representation of the core.
    """

    core_ = core.copy()
    if exchanged and core_.size > 4:
        core_ = core[:, [1, 0], :, :]
    cores = [torch.tensor(core_)] * length
    return TensorTrain(cores)


def concat_ttmaps(
    tt_left: TensorTrain, tt_right: TensorTrain
) -> Tuple[TensorTrain, TensorTrain, TensorTrain]:
    """
    Put together the two TT-map primitives to form the full connectivity maps.
    See Section 5 of arXiv:1802.02839 for details.
    The connectivity maps are defined as follows:
    Pmp describes which nodes in the left domain are connected to which nodes in the right domain,
    Pmm describes which nodes in the left domain are to be connected,
    Ppp describes which nodes in the right domain are to be connected.

    Args:
        tt_left (TensorTrain): The left TT-map.
        tt_right (TensorTrain): The right TT-map.

    Returns:
        Tuple[TensorTrain, TensorTrain, TensorTrain]: The connectivity maps.
    """

    connect_tt = tt_left.t() @ tt_right
    count_left = -(tt_left.t() @ tt_left)
    count_right = -(tt_right.t() @ tt_right)

    return connect_tt, count_left, count_right


def side_concatenation_tt(
    side0: BoundarySide2D, side1: BoundarySide2D, length: int
) -> Tuple[TensorTrain, TensorTrain, TensorTrain]:
    """
    Get the TT-representation of the concatenation tensors of two boundary sides.
    See Section 5 of arXiv:1802.02839 for details.
    The connectivity maps are defined as follows:
    Pmp describes which nodes in the left domain are connected to which nodes in the right domain,
    Pmm describes which nodes in the left domain are to be connected,
    Ppp describes which nodes in the right domain are to be connected.

    Args:
        side0 (BoundarySide2D): The first boundary side.
        side1 (BoundarySide2D): The second boundary side.
        length (int): The length of the TT.

    Returns:
        Tuple[TensorTrain, TensorTrain, TensorTrain]: The connectivity maps.
    """

    core0 = side_concatenation_core(side0)
    core1 = side_concatenation_core(side1)

    tt_left = concat_core2tt(core0, length, exchanged=False)
    tt_right = concat_core2tt(core1, length, exchanged=True)

    return concat_ttmaps(tt_left, tt_right)


def vertex_concatenation_tt(
    vertex0: BoundaryVertex2D, vertex1: BoundaryVertex2D, length: int
) -> Tuple[TensorTrain, TensorTrain, TensorTrain]:
    """
    Get the TT-representation of the concatenation tensor of two boundary vertices.
    See Section 5 of arXiv:1802.02839 for details.
    The connectivity maps are defined as follows:
    Pmp describes which nodes in the left domain are connected to which nodes in the right domain,
    Pmm describes which nodes in the left domain are to be connected,
    Ppp describes which nodes in the right domain are to be connected.

    Args:
        vertex0 (BoundaryVertex2D): The first boundary vertex.
        vertex1 (BoundaryVertex2D): The second boundary vertex.
        length (int): The length of the TT.

    Returns:
        Tuple[TensorTrain, TensorTrain, TensorTrain]: The connectivity maps.
    """

    core0 = vertex_concatenation_core(vertex0)
    core1 = vertex_concatenation_core(vertex1)

    tt_left = concat_core2tt(core0, length, exchanged=False)
    tt_right = concat_core2tt(core1, length, exchanged=False)

    return concat_ttmaps(tt_left, tt_right)



from typing import Tuple

import numpy as np
import torch

from ttfemesh.types import TensorTrain


def left_corner2index_ttmap(mesh_size_exponent: int) -> TensorTrain:
    """
    Returns the TT-representation of the left corner element index to global basis index map (W_0).
    See arXiv:1802.02839, Section 5.4 for more details.
    Note the size of the TT alone already implies we are working with a linear 2D basis.

    Args:
        mesh_size_exponent (int): Exponent of the 1D mesh size (length of TT).

    Returns:
        TensorTrain: TT-representation of the left corner to global index map.
    """
    firstcore, middlecore, lastcore = left_corner2index_map_ttcores()
    cores = [firstcore] + [middlecore] * (mesh_size_exponent - 2) + [lastcore]
    torch_cores = [torch.tensor(core) for core in cores]

    return TensorTrain(torch_cores)


def right_corner2index_ttmap(mesh_size_exponent: int) -> TensorTrain:
    """
    Returns the TT-representation of the right corner element index to global basis index map (W_1).
    See arXiv:1802.02839, Section 5.4 for more details.
    Note the size of the TT alone already implies we are working with a linear 2D basis.

    Args:
        mesh_size_exponent (int): Exponent of the 1D mesh size (length of TT).

    Returns:
        TensorTrain: TT-representation of the right corner to global index map
    """
    firstcore, middlecore, lastcore = right_corner2index_map_ttcores()
    cores = [firstcore] + [middlecore] * (mesh_size_exponent - 2) + [lastcore]
    torch_cores = [torch.tensor(core) for core in cores]

    return TensorTrain(torch_cores)


def left_corner2index_map_ttcores() -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Returns the TT-cores for the left corner element index to global basis index map (W_0).
    See arXiv:1802.02839, Section 5.4 for more details.
    We reverse the order of the cores for consistency with the little-endian
    convention in the QTT literature, i.e.,
    column-major or Fortran-style ordering.

    Returns:
        Tuple[np.ndarray, np.ndarray, np.ndarray]: TT-cores of the left corner to global index map.
    """

    firstcore = np.zeros((1, 2, 2, 2))
    firstcore[0, :, :, 0] = np.array([[1.0, 0.0], [0.0, 1.0]])
    firstcore[0, :, :, 1] = np.array([[1.0, 0.0], [0.0, 0.0]])

    middlecore = np.zeros((2, 2, 2, 2))
    middlecore[0, :, :, 0] = np.array([[1.0, 0.0], [0.0, 1.0]])
    middlecore[0, :, :, 1] = np.array([[1.0, 0.0], [0.0, 0.0]])
    middlecore[1, :, :, 0] = np.array([[0.0, 0.0], [0.0, 0.0]])
    middlecore[1, :, :, 1] = np.array([[0.0, 0.0], [0.0, 1.0]])

    lastcore = np.zeros((2, 2, 2, 1))
    lastcore[0, :, :, 0] = np.array([[1.0, 0.0], [0.0, 0.0]])
    lastcore[1, :, :, 0] = np.array([[0.0, 0.0], [0.0, 1.0]])

    cores = firstcore, middlecore, lastcore
    return cores


def right_corner2index_map_ttcores() -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Returns the TT-cores for the right corner element index to global basis index map (W_1).
    See arXiv:1802.02839, Section 5.4 for more details.
    We reverse the order of the cores for consistency with the little-endian
    convention in the QTT literature, i.e.,
    column-major or Fortran-style ordering.

    Returns:
        Tuple[np.ndarray, np.ndarray, np.ndarray]: TT-cores of the right corner to global index map.
    """
    firstcore = np.zeros((1, 2, 2, 2))
    firstcore[0, :, :, 0] = np.array([[0.0, 1.0], [0.0, 0.0]])
    firstcore[0, :, :, 1] = np.array([[0.0, 0.0], [1.0, 0.0]])

    middlecore = np.zeros((2, 2, 2, 2))
    middlecore[0, :, :, 0] = np.array([[0.0, 1.0], [0.0, 0.0]])
    middlecore[0, :, :, 1] = np.array([[0.0, 0.0], [1.0, 0.0]])
    middlecore[1, :, :, 0] = np.array([[0.0, 0.0], [0.0, 0.0]])
    middlecore[1, :, :, 1] = np.array([[1.0, 0.0], [0.0, 1.0]])

    lastcore = np.zeros((2, 2, 2, 1))
    lastcore[0, :, :, 0] = np.array([[0.0, 0.0], [1.0, 0.0]])
    lastcore[1, :, :, 0] = np.array([[1.0, 0.0], [0.0, 1.0]])

    cores = firstcore, middlecore, lastcore
    return cores



from abc import ABC, abstractmethod
from typing import List, Literal, Tuple

import numpy as np

from ttfemesh.domain.curve import Curve
from ttfemesh.domain.subdomain import Subdomain2D

CurvePosition = Literal["start", "end"]


class SubdomainConnection(ABC):
    """Generic subdomain connection class."""

    @abstractmethod
    def validate(self, *args, **kwargs):  # pragma: no cover
        """
        Validates that the connection is consistent with the provided subdomains.
        """
        pass

    @property
    @abstractmethod
    def dimension(self) -> int:  # pragma: no cover
        """Dimension of the subdomain connection."""
        pass

    @property
    @abstractmethod
    def num_connected_subdomains(self) -> int:  # pragma: no cover
        """Number of connected subdomains."""
        pass


class SubdomainConnection2D(SubdomainConnection):
    """Generic 2D subdomain connection class."""

    @property
    def dimension(self) -> int:
        """Dimension of the subdomain connection."""
        return 2


class VertexConnection2D(SubdomainConnection2D):
    """
    Initialize a 2D vertex connection.
    The subdomain indexes reference into the list of subdomains
    passed to the Domain constructor.

    Args:
        connection (List[Tuple[int, int, CurvePosition]]):
            List of subdomains sharing this vertex.
            Each connection is a tuple of (subdomain index, curve index, position).
            Curve position is either "start" or "end".

    Example:
    >>> from ttfemesh.domain import RectangleFactory, CurveConnection2D, VertexConnection2D
    >>> from ttfemesh.domain import Domain2D
    >>> lower_left = (0, 0)
    >>> upper_right = (2, 1)
    >>> rectangle1 = RectangleFactory.create(lower_left, upper_right)

    >>> lower_left = (2, 0)
    >>> upper_right = (3, 1)
    >>> rectangle2 = RectangleFactory.create(lower_left, upper_right)

    >>> lower_left = (-2, 1)
    >>> upper_right = (0, 2)
    >>> rectangle3 = RectangleFactory.create(lower_left, upper_right)

    >>> domain_idxs = [0, 1]
    >>> curve_idxs = [1, 3]
    >>> edge = CurveConnection2D(domain_idxs, curve_idxs)

    >>> vertex_idxs = [(0, 3, "start"), (2, 0, "end")]
    >>> vertex = VertexConnection2D(vertex_idxs)

    >>> domain = Domain2D([rectangle1, rectangle2, rectangle3], [edge, vertex])
    >>> domain.plot()
    """

    def __init__(self, connection: List[Tuple[int, int, CurvePosition]]):
        self.connection = connection

    @property
    def num_connected_subdomains(self) -> int:
        """Number of connected subdomains."""
        return len(self.connection)

    def validate(self, subdomains: List[Subdomain2D], tol: float = 1e-6):
        """
        Validate that all specified subdomains, curves, and positions share the given vertex.

        Args:
            subdomains (List[Subdomain2D]): List of subdomains in the domain.
            tol (float): Tolerance for point-wise comparison

        Raises:
            ValueError: If the vertex connections are not consistent.
        """
        self._validate_idxs(subdomains)
        if self.num_connected_subdomains < 2:
            raise ValueError("Vertex connection must have at least two connected subdomains.")

        curve0 = subdomains[self.connection[0][0]].curves[self.connection[0][1]]
        point0 = curve0.get_start() if self.connection[0][2] == "start" else curve0.get_end()
        for subdomain_idx, curve_idx, position in self.connection:
            curve = subdomains[subdomain_idx].curves[curve_idx]
            point = curve.get_start() if position == "start" else curve.get_end()

            if not np.allclose(point, point0, atol=tol):
                raise ValueError(
                    f"Subdomain {subdomain_idx}, curve {curve_idx}, {position} point {point} "
                    f"does not match the vertex {point0}."
                )

    def get_connection_pairs(
        self,
    ) -> List[Tuple[Tuple[int, int], Tuple[int, int], Tuple[CurvePosition, CurvePosition]]]:
        """
        Get all unique pairs of connected subdomains with their curve indices and positions.

        Returns:
            List[Tuple[Tuple[int, int], Tuple[int, int], Tuple[CurvePosition, CurvePosition]]]:
                List of connected subdomain pairs with the indexing
                [(subdomain0, subdomain1), (curve0, curve1), (position0, position1)].
        """
        pairs = []
        n = len(self.connection)

        for i in range(n):
            for j in range(i + 1, n):
                subd1, curve_idx1, curve_pos1 = self.connection[i]
                subd2, curve_idx2, curve_pos2 = self.connection[j]
                pairs.append(((subd1, subd2), (curve_idx1, curve_idx2), (curve_pos1, curve_pos2)))

        return pairs

    def get_shared_vertex(self, subdomains: List[Subdomain2D]) -> np.ndarray:
        """
        Get the shared vertex between the connected subdomains.

        Args:
            subdomains (List[Subdomain2D]): List of subdomains in the domain.

        Returns:
            np.ndarray: Shared vertex coordinates.
        """
        self._validate_idxs(subdomains)

        curve0 = subdomains[self.connection[0][0]].curves[self.connection[0][1]]
        return curve0.get_start() if self.connection[0][2] == "start" else curve0.get_end()

    def _validate_idxs(self, subdomains: List[Subdomain2D]):
        """Validate that the subdomain and curve indices are within bounds."""
        for subdomain_idx, curve_idx, position in self.connection:
            if subdomain_idx >= len(subdomains):
                raise ValueError(f"Subdomain index {subdomain_idx} is out of bounds.")
            if curve_idx >= len(subdomains[subdomain_idx].curves):
                raise ValueError(f"Curve index {curve_idx} is out of bounds.")

    def __repr__(self):
        return f"VertexConnection2D({self.connection})"


class CurveConnection2D(SubdomainConnection2D):
    """
    Initialize a curve connection between two subdomains.
    Only two subdomains can be connected by a curve.

    Args:
        subdomains_indices (Tuple[int, int]):
            A tuple of two subdomain indices that share a curve.
        curve_indices (Tuple[int, int]):
            A tuple of two curve indices in the respective subdomains.

    Example:
    >>> from ttfemesh.domain import RectangleFactory, CurveConnection2D, VertexConnection2D
    >>> from ttfemesh.domain import Domain2D
    >>> lower_left = (0, 0)
    >>> upper_right = (2, 1)
    >>> rectangle1 = RectangleFactory.create(lower_left, upper_right)

    >>> lower_left = (2, 0)
    >>> upper_right = (3, 1)
    >>> rectangle2 = RectangleFactory.create(lower_left, upper_right)

    >>> lower_left = (-2, 1)
    >>> upper_right = (0, 2)
    >>> rectangle3 = RectangleFactory.create(lower_left, upper_right)

    >>> domain_idxs = [0, 1]
    >>> curve_idxs = [1, 3]
    >>> edge = CurveConnection2D(domain_idxs, curve_idxs)

    >>> vertex_idxs = [(0, 3, "start"), (2, 0, "end")]
    >>> vertex = VertexConnection2D(vertex_idxs)

    >>> domain = Domain2D([rectangle1, rectangle2, rectangle3], [edge, vertex])
    >>> domain.plot()
    """

    def __init__(self, subdomains_indices: Tuple[int, int], curve_indices: Tuple[int, int]):
        self.subdomains_indices = subdomains_indices
        self.curve_indices = curve_indices

    @property
    def num_connected_subdomains(self) -> int:
        """Number of connected subdomains."""
        return 2

    def validate(self, subdomains: List[Subdomain2D], num_points: int = 100, tol: float = 1e-6):
        """
        Validate that the curves are approximately equal.

        Args:
            subdomains (List[Subdomain2D]): List of subdomains in the domain.
            num_points (int): Number of points to sample along the curve.
            tol (float): Tolerance for point-wise comparison.
        """
        self._validate_idxs(subdomains)

        sub1_idx, sub2_idx = self.subdomains_indices
        curve1_idx, curve2_idx = self.curve_indices

        curve1 = subdomains[sub1_idx].curves[curve1_idx]
        curve2 = subdomains[sub2_idx].curves[curve2_idx]

        if not curve1.equals(curve2, num_points=num_points, tol=tol):
            raise ValueError(
                f"Curves {curve1_idx} of subdomain {sub1_idx}"
                f" and {curve2_idx} of subdomain {sub2_idx} are not equal."
            )

    def get_shared_curve(self, subdomains: List[Subdomain2D]) -> Curve:
        """
        Get the shared curve between the connected subdomains.

        Args:
            subdomains (List[Subdomain2D]): List of subdomains in the domain.

        Returns:
            Curve: Shared curve.
        """
        self._validate_idxs(subdomains)

        sub1_idx, sub2_idx = self.subdomains_indices
        curve1_idx, curve2_idx = self.curve_indices

        return subdomains[sub1_idx].curves[curve1_idx]

    def _validate_idxs(self, subdomains: List[Subdomain2D]):
        """Validate that the subdomain and curve indices are within bounds."""
        sub1_idx, sub2_idx = self.subdomains_indices
        curve1_idx, curve2_idx = self.curve_indices

        if sub1_idx >= len(subdomains):
            raise ValueError(f"Subdomain index {sub1_idx} is out of bounds.")
        if sub2_idx >= len(subdomains):
            raise ValueError(f"Subdomain index {sub2_idx} is out of bounds.")
        if curve1_idx >= len(subdomains[sub1_idx].curves):
            raise ValueError(f"Curve index {curve1_idx} is out of bounds.")
        if curve2_idx >= len(subdomains[sub2_idx].curves):
            raise ValueError(f"Curve index {curve2_idx} is out of bounds.")

    def __repr__(self):
        return f"CurveConnection({self.subdomains_indices}, {self.curve_indices})"
